from typing import Any, List

import torch

from esp_ppq.core import DataType, convert_any_to_python_primary_type, convert_any_to_torch_tensor, ppq_warning
from esp_ppq.IR.search import SearchableGraph

from .base.command import (
    GraphCommand,
    GraphCommandType,
    ReplaceOperationCommand,
    ReplaceVariableCommand,
    TruncateGraphCommand,
)
from .base.graph import Operation, Variable
from .processer import GraphCommandProcessor


class GraphReplacer(GraphCommandProcessor):
    """Graph Replacer offers a bunch of graph editing functions that helps replacing operation or variable in your graph."""

    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.REPLACE_OP:
            assert isinstance(command, ReplaceOperationCommand), 'Use ReplaceOperationCommand instead of GraphCommand'
            return self.replace_op(command.op_name, command.replace_to)
        if command.command_type == GraphCommandType.REPLACE_VAR:
            assert isinstance(command, ReplaceVariableCommand), 'Use ReplaceOperationCommand instead of GraphCommand'
            return self.replace_var(command.op_name, command.replace_to)
        if command.command_type == GraphCommandType.REPLACE_BATCHNORM_TO_CONV:
            return self.replace_batchnorm_to_conv()
        if command.command_type == GraphCommandType.REPLACE_BATCHNORM_TO_SCALE:
            return self.replace_batchnorm_to_scale()

    def replace_op(self, op_name: str, replace_to: Operation):
        if op_name not in self._graph.operations:
            raise KeyError(f'Operation {op_name} is not in current graph')
        operation = self._graph.operations[op_name]

        replace_to.inputs.clear()
        replace_to.inputs.extend(operation.inputs)
        for input_var in operation.inputs:
            dest_idx = input_var.dest_ops.index(operation)
            input_var.dest_ops[dest_idx] = replace_to

        replace_to.outputs.clear()
        replace_to.outputs.extend(operation.outputs)
        for output_var in operation.outputs:
            output_var.source_op = replace_to

        replace_to.parameters.clear()
        replace_to.parameters.extend(operation.parameters)

        self._graph.operations[op_name] = replace_to

    def replace_var(self, var_name: str, replace_to: Variable):
        if var_name not in self._graph.variables:
            raise KeyError(f'Variable {var_name} is not in current graph')
        variable = self._graph.variables[var_name]

        replace_to.dest_ops.clear()
        replace_to.dest_ops.extend(variable.dest_ops)
        for dest_op in replace_to.dest_ops:
            dest_idx = dest_op.inputs.index(variable)
            dest_op.inputs[dest_idx] = replace_to

        replace_to.source_op = variable.source_op
        if variable.source_op is not None:
            source_idx = variable.source_op.outputs.index(variable)
            variable.source_op.outputs[source_idx] = replace_to

        self._graph.variables[var_name] = replace_to
        if var_name in self._graph.inputs:
            self._graph.inputs[var_name] = replace_to
        if var_name in self._graph.outputs:
            self._graph.outputs[var_name] = replace_to

    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [
            GraphCommandType.REPLACE_OP,
            GraphCommandType.REPLACE_VAR,
            GraphCommandType.REPLACE_BATCHNORM_TO_CONV,
            GraphCommandType.REPLACE_BATCHNORM_TO_SCALE,
        ]

    def replace_batchnorm_to_conv(self, dimension: int = 2):
        """Replace Batchnorm to 1D Convolution."""
        for op in self.graph.operations.values():
            if op.type == 'BatchNormalization':
                ppq_warning(
                    f'Isolated BatchNormalization({op.name}) was detected, '
                    f'PPQ will replace it to 1*1 Convolution({dimension}D).'
                )

                assert len(op.parameters) == 4, "BatchNorm should have 4 parameters, namely alpha, beta, mean, var"
                alpha = op.parameters[0].value
                beta = op.parameters[1].value
                mean = op.parameters[2].value
                var = op.parameters[3].value
                epsilon = op.attributes.get("epsilon", 1e-5)

                with torch.no_grad():
                    w = alpha / torch.sqrt(var + epsilon)
                    w = w.reshape([-1, 1] + [1] * dimension)
                    b = alpha * (-mean) / torch.sqrt(var + epsilon) + beta

                op.type = 'Conv'
                op.attributes.clear()
                op.attributes['kernel_shape'] = [1] * dimension
                op.attributes['strides'] = [1] * dimension
                op.attributes['dilations'] = [1] * dimension
                op.attributes['pads'] = [0, 0] * dimension
                op.attributes['group'] = w.numel()

                # remove last 2 variable, make conv has exact 3 input
                self.graph.remove_variable(op.inputs[-1])
                self.graph.remove_variable(op.inputs[-1])

                with torch.no_grad():
                    op.inputs[1].value = w
                    op.inputs[2].value = b

    def replace_batchnorm_to_scale(self, dimension: int = 4):
        """Replace Batchnorm to Mul + Add.

        By default this function created a 4d mul + add corresponding to NCHW layout.
        """
        graph = self.graph
        for op in [_ for _ in self.graph.operations.values()]:
            if op.type == 'BatchNormalization':
                ppq_warning(
                    f'Isolated BatchNormalization({op.name}) was detected, '
                    f'PPQ will replace it to Mul + Add({dimension}D).'
                )

                assert len(op.parameters) == 4, "BatchNorm should have 4 parameters, namely alpha, beta, mean, var"
                alpha = op.parameters[0].value
                beta = op.parameters[1].value
                mean = op.parameters[2].value
                var = op.parameters[3].value
                epsilon = op.attributes.get("epsilon", 1e-5)

                with torch.no_grad():
                    multiplier = alpha / torch.sqrt(var + epsilon)
                    bias = (-mean) * multiplier + beta

                for var in [_ for _ in op.parameters]:
                    graph.remove_variable(var)
                graph.create_variable(value=multiplier, is_parameter=True, dest_ops=[op])
                op.type = 'Mul'
                op.attributes.clear()

                add = graph.create_operation(op_type='Add')
                graph.insert_op_after(A=add, B=op)
                graph.create_variable(value=bias, is_parameter=True, dest_ops=[add])

                if dimension > 1:
                    op.parameters[0].value = op.parameters[0].value.reshape([1, -1] + [1] * (dimension - 2))
                    add.parameters[0].value = add.parameters[0].value.reshape([1, -1] + [1] * (dimension - 2))


class GraphFormatter(GraphCommandProcessor):
    """Graph Formatter offers a bunch of graph editing functions that helps modifying your graph."""

    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [
            GraphCommandType.FORMAT_CLIP,
            GraphCommandType.FORMAT_PAD,
            GraphCommandType.FORMAT_GATHER,
            GraphCommandType.FORMAT_CAST,
            GraphCommandType.FORMAT_INT64_CONSTANT,
            GraphCommandType.DELETE_ISOLATED,
            GraphCommandType.FORMAT_PARAMETERS,
            GraphCommandType.FORMAT_CONSTANT_INPUT,
            GraphCommandType.FORMAT_SLICE,
            GraphCommandType.TRUNCATE_ON_VAR,
            GraphCommandType.FORMAT_RESIZE,
            GraphCommandType.REMOVE_IDENTITY,
            GraphCommandType.CONVERT_TO_TENSOR,
        ]

    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.FORMAT_CLIP:
            return self.format_clip()
        if command.command_type == GraphCommandType.FORMAT_PAD:
            return self.format_pad()
        if command.command_type == GraphCommandType.FORMAT_GATHER:
            return self.format_gather()
        if command.command_type == GraphCommandType.FORMAT_CAST:
            return self.format_cast()
        if command.command_type == GraphCommandType.DELETE_ISOLATED:
            return self.delete_isolated()
        if command.command_type == GraphCommandType.FORMAT_INT64_CONSTANT:
            return self.format_int64_constant()
        if command.command_type == GraphCommandType.FORMAT_PARAMETERS:
            return self.format_parameter()
        if command.command_type == GraphCommandType.FORMAT_CONSTANT_INPUT:
            return self.remove_constant_input()
        if command.command_type == GraphCommandType.FORMAT_SLICE:
            return self.format_slice()
        if command.command_type == GraphCommandType.FORMAT_RESIZE:
            return self.format_resize()
        if command.command_type == GraphCommandType.TRUNCATE_ON_VAR:
            assert isinstance(command, TruncateGraphCommand), 'Use TruncateGraphCommand here.'
            return self.truncate_on_var(command.var, command.mark_as_output)
        if command.command_type == GraphCommandType.REMOVE_IDENTITY:
            return self.remove_identity()
        if command.command_type == GraphCommandType.CONVERT_TO_TENSOR:
            return self.convert_to_tensor()

    def format_slice(self) -> None:
        """
        Slice: opset1 格式跟其他的不太一样，这个 pass 将 opset1 的 slice 强行转换为 opset 11
        """
        interested_ops = []
        for operation in self.graph.operations.values():
            if operation.type == 'Slice':
                if 'starts' in operation.attributes:
                    assert 'starts' in operation.attributes and 'ends' in operation.attributes, (
                        f'Invalid Slice Operation Format, Slice operation is expected to have axes, '
                        'starts and ends attributes with opset 1, '
                        f'however your operation {operation.name}, do not have completed attributes'
                    )
                    interested_ops.append(operation)

        for slice in interested_ops:
            assert isinstance(slice, Operation)
            axes = slice.attributes.get('axes', None)
            starts = slice.attributes['starts']
            ends = slice.attributes['ends']
            if axes == None:
                axes = [_ for _ in range(starts)]

            slice.attributes.pop('starts')
            slice.attributes.pop('ends')
            if 'axes' in slice.attributes:
                slice.attributes.pop('axes')
            self.__add_constant_input(slice, convert_any_to_torch_tensor(starts))
            self.__add_constant_input(slice, convert_any_to_torch_tensor(ends))
            self.__add_constant_input(slice, convert_any_to_torch_tensor(axes))

    def format_pad(self) -> None:
        """
        对于不同的模型格式, pad 算子将有两种不同的输入格式：
        for different models, possibly Pad op has the following input formats
            1. pads 参数由第二个输入变量给出
               pads parameter is given by the second input variable
            2. pads 参数被放置于 operation.attribute 中
               pads parameter is set in attribute
        此函数统一 pad 算子行为：所有 pad 算子的 pads 参数均由第二个输入给出
        """
        for op in self.graph.operations.values():
            if op.type == 'Pad' and 'pads' in op.attributes:
                self.graph.create_variable(value=torch.tensor(op.attributes['pads']), is_parameter=True, dest_ops=[op])
                op.attributes.clear()

    def format_resize(self) -> None:
        """
        升级 opset 10 的 resize 到 opset 11
        """
        for op in self.graph.operations.values():
            if op.type == 'Resize' and len(op.inputs) == 2:
                # 创建一个空的 variable, 这个 variable 没有值与 source_op, 也不是 parameter
                # 这种行为 Onnx 是允许的，这种变量用以传递空值，起到占位符的作用
                self.graph.create_variable(value=None, is_parameter=False, dest_ops=[op])
                op.inputs[1], op.inputs[2] = op.inputs[2], op.inputs[1]

    def format_clip(self) -> None:
        """
        对于不同的模型格式, clip 算子将有两种不同的输入格式：
        for different models, possibly clip op has the following input formats
            1. min, max 参数由 第二、第三个输入变量给出
               min, max parameter will be given by the second and third input variable
            2. min, max 参数由 attribute 给出
               min, max parameter will be given by the attribute
        此函数统一 clip 算子行为：所有 clip 算子的 min, max 参数第二第三个变量给出
        this func unifies behaviors of clip op: min, max parameter will be given by input vars
        针对可能存在的 min, max 为空的情况，将其直接置为 2 << 30（保证处理后非空）

        当 min, max 参数由 第二、第三个输入变量给出时，其中一个为空时直接返回 ValueError
        ValueError will be raised when any of min, max parameters is null
        """

        interested_ops = []
        for _, operation in self.graph.operations.items():
            if operation.type == 'Clip' and ('min' in operation.attributes or 'max' in operation.attributes):
                interested_ops.append(operation)
        for op in interested_ops:
            assert isinstance(op, Operation)
            min = op.attributes.get('min', -2 << 30)
            max = op.attributes.get('max', +2 << 30)
            min_var = Variable(name=op.name + '_min', value=min, is_parameter=True, dest_ops=[op])
            max_var = Variable(name=op.name + '_max', value=max, is_parameter=True, dest_ops=[op])
            self.graph.append_variable(min_var)
            self.graph.append_variable(max_var)
            op.inputs.append(min_var)
            op.inputs.append(max_var)
            if 'min' in op.attributes:
                op.attributes.pop('min')
            if 'max' in op.attributes:
                op.attributes.pop('max')

    def format_gather(self) -> None:
        """gather op 的参数 index 可能由 input variable 给出 但 index
        参数不可以被量化，同时后端运算需要其作为Python 原生类型 因此将其转移到 gather op 的属性上。 index parameter
        of gather op can be given by input variable, however, it can't be
        quantized, thus we transfer index parameter to attribute of gather op.

        axis is set to 0 when it's not given gather op 的参数 axis 可能不存在，此时强制植入 0
        作为 axis 参数
        """
        interested_ops = []
        for _, operation in self.graph.operations.items():
            if operation.type == 'Gather':
                interested_ops.append(operation)
        for operation in interested_ops:
            assert isinstance(operation, Operation)
            if len(operation.inputs) == 2:
                index_op = operation.inputs[1].source_op
                if index_op.type == 'Constant':
                    index = index_op.attributes['value']
                    self.__delete_constant_input(operation, 1)
                    operation.attributes['gather_index'] = convert_any_to_python_primary_type(index)
            if 'axis' not in operation.attributes:
                operation.attributes['axis'] = 0

            if 'indices' in operation.attributes:
                operation.attributes['gather_index'] = operation.attributes['indices']
                operation.attributes.pop('indices')

    def format_cast(self) -> None:
        """cast op 的参数 to 默认为 int，使用该函数将其封装为 esp_ppq.core.DataType."""
        interested_ops = []
        for _, operation in self.graph.operations.items():
            assert isinstance(operation, Operation)
            if operation.type == 'Cast':
                interested_ops.append(operation)
        for operation in interested_ops:
            assert isinstance(operation, Operation)
            assert 'to' in operation.attributes
            operation.attributes['to'] = DataType.convert_from_numpy(operation.attributes['to'])

    def format_int64_constant(self) -> None:
        """convert all int64 constants to int32, check if direct dtype cast is
        available 将所有 int64 的 Constant 转换为 int32 将检查所有 Constant value, 如果 value
        范围在 int32 表示范围内则执行转换。"""
        for operation in self.graph.operations.values():
            if operation.type == 'Constant':
                assert 'value' in operation.attributes
                value = operation.attributes['value']

                assert isinstance(value, torch.Tensor)
                if value.dtype != torch.int64:
                    continue

                pvalue = convert_any_to_python_primary_type(value)
                check = [0xFFFFFFFF > v >= -0xFFFFFFFF for v in pvalue]

                if all(check):
                    value = value.int()

    def remove_constant_input(self) -> None:
        """部分部署平台不支持 Constant Op 作为算子的输入
            在这种情况下我们使用这个 pass 把它们切换成 Parameter Variable

        Some backend platform doesn't support Constant
        Op, we use this pass to replace it by forcing its value to be a
        parameter variable."""
        removing_ops = []
        for op in self.graph.operations.values():
            if op.type == 'Constant':
                assert len(op.outputs) == 1, (
                    f'Constant Operation {op.name} has more than 1 output, is there a network parsing error?'
                )
                removing_ops.append(op)

        for const_op in removing_ops:
            assert isinstance(const_op, Operation)
            constant_value = const_op.attributes['value']
            output_var = const_op.outputs[0]
            output_var._is_parameter = True
            output_var.value = constant_value
            self.graph.remove_operation(removing_op=const_op)

    def truncate_on_var(self, var: Variable, mark_as_output: bool):
        """从一个指定位置将图截断.

        Args:
            var (Variable): _description_
            mark_as_output (bool): _description_

        Raises:
            TypeError: _description_
            KeyError: _description_
        """
        graph = self.graph
        if not isinstance(var, Variable):
            raise TypeError(f'Except variable instance here, however {type(var)} was given.')
        if var.name not in graph.variables:
            raise KeyError(f'Can not find vairiable {var.name} in current graph')

        mark_to_delete, delete_queue, didx = set(), [], 0
        delete_queue.extend(var.dest_ops)
        while didx < len(delete_queue):
            first_op = delete_queue[didx]
            if first_op not in mark_to_delete:
                mark_to_delete.add(first_op)
                delete_queue.extend(graph.get_downstream_operations(first_op))
            didx += 1

        for operation in mark_to_delete:
            graph.remove_operation(operation)

        if mark_as_output:
            graph.mark_variable_as_graph_output(var)

        self.delete_isolated()

    def delete_isolated(self):
        """Remove Isolated Variable from Graph."""
        blacklist = [None]
        while len(blacklist) > 0:
            blacklist = []
            # delete all operations which are not links to a valid graph output
            for op in self.graph.operations.values():
                if len(self.graph.get_downstream_operations(op)) == 0:
                    output_names = [var.name for var in op.outputs]
                    if all([name not in self.graph.outputs for name in output_names]):
                        blacklist.append(op)

            for op in blacklist:
                for var in op.outputs:
                    self.graph.remove_variable(var)
                self.graph.remove_operation(op)

        var_blacklist = [None]
        while len(var_blacklist) > 0:
            var_blacklist = set()
            # delete all variables that links to invalid operations:
            for var in self.graph.variables.values():
                # 删除无根无输出的变量
                if var.source_op is None and len(var.dest_ops) == 0:
                    var_blacklist.add(var)

                # 删除根节点不在图中的变量
                if var.source_op is not None and var.source_op.name not in self.graph.operations:
                    var_blacklist.add(var)

                # 删除连接到未知节点的变量
                for op in var.dest_ops:
                    if op.name not in self.graph.operations:
                        var_blacklist.add(var)

                # 删除孤立变量
                if var.source_op is None and var.name not in self.graph.inputs:
                    if len(var.dest_ops) == 0:
                        var_blacklist.add(var)

                # 没有输出的不能删...会影响算子输出顺序...

            for var in var_blacklist:
                self.graph.remove_variable(var)

    def format_parameter(self) -> None:
        """Split parameter that has more than 1 dest ops"""
        for var in [_ for _ in self.graph.variables.values()]:
            var.value = convert_any_to_torch_tensor(var.value)
            if var.is_parameter and len(var.dest_ops) > 1:
                for op in var.dest_ops:
                    created = self.graph.create_variable(value=var.value.clone(), is_parameter=True)
                    op.inputs[op.inputs.index(var)] = created
                    created.dest_ops.append(op)
                var.dest_ops.clear()
                self.graph.remove_variable(var)

    def __delete_constant_input(self, op: Operation, input_idx: int):
        op_name = op.name
        if op_name not in self._graph.operations:
            raise KeyError(f'Operation {op_name} not in current graph.')
        operation = self._graph.operations[op_name]
        assert input_idx < len(operation.inputs), (
            'Trying to delete an out-of-range input variable, '
            f'has graph been manually changed? Error at Operation {op_name}, input_idx: {input_idx}'
        )
        input_var = operation.inputs[input_idx]
        if input_var.source_op.type != 'Constant':
            raise ValueError(f'Trying to delete an non-const input, Error at Operation {op_name}, inputs[{input_idx}]')
        input_var.dest_ops.pop(input_var.dest_ops.index(operation))
        operation.inputs.pop(input_idx)

        if len(input_var.dest_ops) == 0:
            self.graph.remove_operation(input_var.source_op)
            self.graph.remove_variable(input_var)

    def __add_constant_input(self, op: Operation, value: torch.Tensor):
        op_name = op.name
        if op_name not in self._graph.operations:
            raise KeyError(f'Operation {op_name} not in current graph.')
        operation = self._graph.operations[op_name]
        var = Variable(name=f'{op_name}_{len(op.inputs) + 1}', value=value, is_parameter=True)
        self.graph.append_variable(var)
        var.dest_ops.append(operation)
        operation.inputs.append(var)

    def remove_identity(self):
        removing_ops = []
        for op in self.graph.operations.values():
            if op.type == 'Identity':
                removing_ops.append(op)

        for op in removing_ops:
            self.graph.remove_operation(op, keep_coherence=True)

    def convert_to_tensor(self):
        """Convert anything inside your network to torch tensor. (Usually from numpy)"""
        for var in self.graph.variables.values():
            if var.value is not None:
                var.value = convert_any_to_torch_tensor(var.value)


class GraphMerger(GraphCommandProcessor):
    """Graph Merger implements all graph fusion related functions."""

    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [
            # add more extensions in the future
            GraphCommandType.FUSE_BN,
            GraphCommandType.FUSE_BIAS_ADD,
            GraphCommandType.FUSE_SWISH,
        ]

    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.FUSE_BN:
            return self.fuse_bn()
        if command.command_type == GraphCommandType.FUSE_BIAS_ADD:
            return self.fuse_bias_add()
        if command.command_type == GraphCommandType.FUSE_SWISH:
            return self.fuse_swish()

    def fuse_bn(self):
        search_engine = SearchableGraph(graph=self.graph)
        paths = search_engine.path_matching(
            sp_expr=lambda x: x.type in {'Conv', 'Gemm', 'ConvTranspose'},
            rp_expr=lambda x, y: False,
            ep_expr=lambda x: x.type == 'BatchNormalization',
            direction='down',
        )

        for path in paths:
            path = path.tolist()
            assert len(path) == 2, 'Oops seems we got something unexpected.'

            computing_op, bn_op = path
            assert isinstance(computing_op, Operation) and isinstance(bn_op, Operation)

            if (
                len(self.graph.get_downstream_operations(computing_op)) != 1
                or len(self.graph.get_upstream_operations(bn_op)) != 1
            ):
                ppq_warning(
                    f'PPQ can not merge operation {computing_op.name} and {bn_op.name}, '
                    'this is not suppose to happen with your network, '
                    'network with batchnorm inside might not be able to quantize and deploy.'
                )
                continue

            assert len(bn_op.parameters) == 4, 'BatchNorm should have 4 parameters, namely alpha, beta, mean, var'
            alpha = bn_op.parameters[0].value
            beta = bn_op.parameters[1].value
            mean = bn_op.parameters[2].value
            var = bn_op.parameters[3].value
            epsilon = bn_op.attributes.get('epsilon', 1e-5)

            if computing_op.num_of_parameter == 1:
                w = computing_op.parameters[0].value  # no bias.
                assert isinstance(w, torch.Tensor), 'values of parameters are assumed as torch Tensor'
                if computing_op.type == 'ConvTranspose':
                    b = torch.zeros(w.shape[1] * computing_op.attributes.get('group', 1))
                elif computing_op.type == 'Gemm' and computing_op.attributes.get('transB', 0) == 0:
                    b = torch.zeros(w.shape[1])
                else:
                    b = torch.zeros(w.shape[0])
            else:
                w, b = [var.value for var in computing_op.parameters[:2]]  # has bias.

            if computing_op.type == 'Conv':
                # calculate new weight and bias
                scale = alpha / torch.sqrt(var + epsilon)
                w = w * scale.reshape([-1] + [1] * (w.ndim - 1))
                b = alpha * (b - mean) / torch.sqrt(var + epsilon) + beta

            elif computing_op.type == 'Gemm':
                # calculate new weight and bias
                scale = alpha / torch.sqrt(var + epsilon)
                if computing_op.attributes.get('transB', 0):
                    w = w * scale.reshape([-1, 1])
                else:
                    w = w * scale.reshape([1, -1])
                b = alpha * (b - mean) / torch.sqrt(var + epsilon) + beta

            elif computing_op.type == 'ConvTranspose':
                scale = alpha / torch.sqrt(var + epsilon)
                group = computing_op.attributes.get('group', 1)
                scale = scale.reshape([group, 1, -1] + [1] * (w.ndim - 2))
                w = w.reshape([group, -1] + list(w.shape[1:])) * scale
                w = w.reshape([w.shape[0] * w.shape[1]] + list(w.shape[2:]))
                b = alpha * (b - mean) / torch.sqrt(var + epsilon) + beta
            else:
                raise TypeError(
                    f'Unexpected op type {computing_op.type}. Can not merge {computing_op.name} with {bn_op.name}'
                )

            # create new op and variable
            merged_op = Operation(
                computing_op.name, op_type=computing_op.type, attributes=computing_op.attributes.copy()
            )
            weight_var = Variable(computing_op.name + '_weight', w, True, [merged_op])
            bias_var = Variable(computing_op.name + '_bias', b, True, [merged_op])

            # replace & dirty work
            input_var = computing_op.inputs[0]
            output_var = bn_op.outputs[0]

            input_var.dest_ops.remove(computing_op)
            input_var.dest_ops.append(merged_op)

            output_var.source_op = merged_op

            # delete old operations
            computing_op.inputs.pop(0)
            bn_op.outputs.clear()
            self.graph.remove_operation(computing_op)
            self.graph.remove_operation(bn_op)

            # insert new
            self.graph.append_operation(merged_op)
            merged_op.inputs.extend([input_var, weight_var, bias_var])
            merged_op.outputs.extend([output_var])

            self.graph.append_variable(weight_var)
            self.graph.append_variable(bias_var)

    def fuse_gemm(self):
        """Fuse MatMul + add into a singal Gemm
            Single Matmul will be replaced with Gemm

        Returns:
            _type_: _description_
        """

        def _is_replaceable(op: Operation) -> bool:
            if op.inputs[0].is_parameter == False and op.inputs[1].is_parameter == False:
                return False
            else:
                return True

        search_engine = SearchableGraph(graph=self.graph)
        patterns = search_engine.pattern_matching(patterns=["MatMul", "Add"], edges=[[0, 1]], exclusive=True)
        for pattern in patterns:
            matmul, add = pattern

            if _is_replaceable(matmul) == False:
                continue

            matmul.type = "Gemm"

            matmul_out = matmul.outputs[0]
            add_out = add.outputs[0]

            if matmul.inputs[0].is_parameter:
                temp = matmul.inputs[0]
                matmul.inputs[0] = matmul.inputs[1]
                matmul.inputs[1] = temp

            assert len(add.inputs) == 2, "Oops, seems we got some problem here."
            var1, var2 = add.inputs
            bias_var = None

            if var1.source_op == matmul and var2.is_parameter:
                bias_var = var2

            if var2.source_op == matmul and var1.is_parameter:
                bias_var = var1

            # can not find a valid bias, just skip add.
            if bias_var is None:
                continue

            if len(bias_var.value.shape) == 1:
                if bias_var.value.shape[0] == matmul.parameters[0].value.shape[-1]:
                    matmul.attributes["transB"] = 1
                    weight_val = matmul.parameters[0].value

                    matmul.parameters[0].value = weight_val.transpose(-1, -2)

                    bias_var.dest_ops.clear()
                    add.inputs.remove(bias_var)

                    # remove bias add, move bias to matmul
                    self.graph.remove_operation(add)
                    self.graph.create_link_with_op(variable=bias_var, A=None, B=matmul)
                    self.graph.create_link_with_var(A=matmul_out, B=add_out)
                elif bias_var.value.shape[0] == matmul.parameters[0].value.shape[-2]:
                    bias_var.dest_ops.clear()
                    add.inputs.remove(bias_var)
                    # remove bias add, move bias to matmul
                    self.graph.remove_operation(add)
                    self.graph.create_link_with_op(variable=bias_var, A=None, B=matmul)
                    self.graph.create_link_with_var(A=matmul_out, B=add_out)

        # process single gemm
        for op in self.graph.operations.values():
            if op.type == "MatMul":
                if _is_replaceable(op) == False:
                    continue
                op.type = "Gemm"

    def fuse_layernorm(self, exclusive_search: bool = False):
        """Fuse Layernormalization with pattern matching."""

        def _fuse(
            rm1: Operation,
            rm2: Operation,
            eps: Operation,
            scale: torch.Tensor,
            bias: torch.Tensor,
            layernorm_input_var: Variable,
            layernorm_output_var: Variable,
        ) -> Operation:
            if rm2.type == rm1.type == 'ReduceMean':
                if 'axes' not in rm1.attributes:
                    return None
                if 'axes' not in rm2.attributes:
                    return None
                if rm1.attributes['axes'] != rm2.attributes['axes']:
                    return None
                layernorm_axis = rm1.attributes['axes']
                if isinstance(layernorm_axis, list):
                    if len(layernorm_axis) != 1:
                        return None
                    layernorm_axis = layernorm_axis[0]
                if not isinstance(layernorm_axis, int):
                    return None
            else:
                layernorm_axis = -1

            if not eps.inputs[-1].is_parameter:
                return None
            value = eps.inputs[-1].value
            value = convert_any_to_torch_tensor(value).cpu()
            if value.numel() != 1:
                return None
            layernorm_eps = value.item()

            layernorm_output_var.source_op.outputs.clear()
            layernorm = self.graph.create_operation(
                op_type='LayerNormalization',
                attributes={'axis': layernorm_axis, 'epsilon': layernorm_eps, 'stash_type': 0},
                inputs=[layernorm_input_var, self.graph.create_variable(value=scale, is_parameter=True)],
                outputs=[layernorm_output_var],
            )

            if bias is not None:
                self.graph.create_link_with_op(
                    variable=self.graph.create_variable(value=bias, is_parameter=True), A=None, B=layernorm
                )
            return layernorm

        search_engine = SearchableGraph(graph=self.graph)
        fused = False

        # pattern 1:
        #                                 ---     ---     ---      ---        ---       ---    ---    --
        #                               |                                                              |
        # ***(0) --- ReduceMean(1) --- Sub(2) --- Pow(3) --- ReduceMean(4) --- Add(5) --- Sqrt(6) --- Div(7) --- Mul(8) --- (Add)(9)
        #      |                     |
        #       ---   ---   ---   ---
        matches = search_engine.pattern_matching(
            patterns=[
                lambda x: True,
                lambda x: x.type in {'ReduceMean', 'GlobalAveragePool'},
                'Sub',
                'Pow',
                lambda x: x.type in {'ReduceMean', 'GlobalAveragePool'},
                'Add',
                'Sqrt',
                'Div',
                'Mul',
            ],
            edges=[[0, 1], [0, 2], [1, 2], [2, 3], [2, 7], [3, 4], [4, 5], [5, 6], [6, 7], [7, 8]],
            exclusive=exclusive_search,
        )

        for _, rm1, sub, pow, rm2, add, sqrt, div, mul in matches:
            layernorm_ops = [rm1, sub, pow, rm2, add, sqrt, div, mul]

            layernorm_scale = mul.inputs[-1].value
            layernorm_output_var = div.outputs[0]
            layernorm_input_var = sub.inputs[0]

            # bias check
            layernorm_bias = None
            next_op = self.graph.get_downstream_operations(mul)
            if len(next_op) == 1 and (next_op[0].type == 'Add'):
                bias_op = next_op[0]
                if bias_op.inputs[-1].is_parameter:
                    layernorm_bias = bias_op.inputs[-1].value
                    layernorm_output_var = bias_op.outputs[0]
                    layernorm_ops.append(bias_op)

            layernorm = _fuse(
                rm1=rm1,
                rm2=rm2,
                eps=add,
                scale=layernorm_scale,
                bias=layernorm_bias,
                layernorm_input_var=layernorm_input_var,
                layernorm_output_var=layernorm_output_var,
            )

            if layernorm is not None:
                # delete merged ops
                for op in layernorm_ops:
                    assert isinstance(op, Operation)
                    for var in op.inputs + op.outputs:
                        if var != layernorm_input_var and var != layernorm_output_var:
                            self.graph.remove_variable(var)
                    self.graph.remove_operation(op)
                    fused = True

        # pattern 2:
        #  ---   ---  ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---  -- Mul(11)   ---   ---   Add(12) ---
        #  ^                             |                                                                                ^                     ^
        #  |                             v                                                                                |                     |
        # ***(0) --- ReduceMean(1) --- Sub(2) --- Mul(3) --- ReduceMean(4) --- Add(5) --- Sqrt(6) --- Reciprocal(7) --- Mul(8) --- Mul(9) --- Sub(10)
        #             |                                                                                                             |
        #             v                                                                                                             ^
        #             --   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---   ---  ---  --- --
        matches = search_engine.pattern_matching(
            patterns=[
                lambda x: True,
                lambda x: x.type in {'ReduceMean', 'GlobalAveragePool'},
                'Sub',
                'Mul',
                lambda x: x.type in {'ReduceMean', 'GlobalAveragePool'},
                'Add',
                'Sqrt',
                'Reciprocal',
                'Mul',
                'Mul',
                'Sub',
                'Mul',
                'Add',
            ],
            edges=[
                [0, 1],
                [0, 2],
                [0, 11],
                [1, 2],
                [1, 9],
                [2, 3],
                [3, 4],
                [4, 5],
                [5, 6],
                [6, 7],
                [7, 8],
                [8, 9],
                [8, 11],
                [9, 10],
                [11, 12],
                [10, 12],
            ],
            exclusive=exclusive_search,
        )

        for _, rm1, sub1, mul, rm2, add1, sqrt, recipro, mul2, mul3, sub2, mul4, add2 in matches:
            layernorm_ops = [rm1, sub1, mul, rm2, add1, sqrt, recipro, mul2, mul3, sub2, mul4, add2]

            # mul check
            if not mul2.inputs[-1].is_parameter:
                continue
            layernorm_scale = mul2.inputs[-1].value
            layernorm_output_var = add2.outputs[0]
            layernorm_input_var = sub1.inputs[0]
            layernorm_bias = sub2.inputs[0].value

            layernorm = _fuse(
                rm1=rm1,
                rm2=rm2,
                eps=add1,
                scale=layernorm_scale,
                bias=layernorm_bias,
                layernorm_input_var=layernorm_input_var,
                layernorm_output_var=layernorm_output_var,
            )

            if layernorm is not None:
                # delete merged ops
                for op in layernorm_ops:
                    assert isinstance(op, Operation)
                    for var in op.inputs + op.outputs:
                        if var != layernorm_input_var and var != layernorm_output_var:
                            self.graph.remove_variable(var)
                    self.graph.remove_operation(op)
                fused = True

        # final check, if no valid pattern was found, we give a warning.
        if not fused:
            ppq_warning('No valid layernorm pattern was found, check your graph again.')

    def fuse_skiplayernorm(self):
        """Fuse Add + Layernorm to SkipLayernorm, SkipLayernorm is a plugin operation defined by TensorRT"""
        fused = False
        search_engine = SearchableGraph(graph=self.graph)

        matches = search_engine.pattern_matching(patterns=['Add', 'LayerNormalization'], edges=[[0, 1]], exclusive=True)
        for add, layernorm in matches:
            # Skip connection can not be constant.
            if add.num_of_parameter == 0:
                input_vars = add.inputs.copy()
                output_var = add.outputs[0]
                self.graph.remove_operation(add)
                self.graph.remove_variable(output_var)

                for var in input_vars:
                    var.dest_ops.append(layernorm)
                    layernorm.inputs.append(var)

                layernorm.type = 'skipLayerNormPlugin'
                fused = True
        # final check, if no valid pattern was found, we give a warning.
        if not fused:
            ppq_warning('No valid Skip Layernorm pattern was found, check your graph again.')

    def fuse_gelu(self):
        """Fuse Gelu

        Pattern: * - Div - Erf - Add - Mul - Mul
                   |                 |
                   -------------------
        """
        fused = False
        search_engine = SearchableGraph(graph=self.graph)

        matches = search_engine.pattern_matching(
            patterns=[lambda x: True, 'Div', 'Erf', 'Add', 'Mul', 'Mul'],
            edges=[[0, 1], [1, 2], [2, 3], [3, 4], [0, 4], [4, 5]],
            exclusive=True,
        )

        for _, div, erf, add, mul1, mul2 in matches:
            removing_var = []
            removing_var.extend(div.outputs)
            removing_var.extend(erf.outputs)
            removing_var.extend(add.outputs)
            removing_var.extend(mul1.outputs)

            self.graph.remove_operation(div)
            self.graph.remove_operation(erf)
            self.graph.remove_operation(add)
            self.graph.remove_operation(mul1)
            for var in removing_var:
                self.graph.remove_variable(var)

            input_vars = _.outputs.copy()
            output_vars = mul2.outputs.copy()

            self.graph.remove_operation(mul2)
            self.graph.create_operation(op_type='Gelu', inputs=input_vars, outputs=output_vars)
            assert len(input_vars) == 1, 'Fusion failed, Pattern unrecognized.'
            fused = True

        # final check, if no valid pattern was found, we give a warning.
        if not fused:
            ppq_warning('No valid Gelu pattern was found, check your graph again.')

    def fuse_swish(self):
        """
        Fuse Pattern like Sigmoid + Mul to Swish.
        """
        search_engine = SearchableGraph(graph=self.graph)
        patterns = search_engine.pattern_matching(
            patterns=[lambda x: x.is_computing_op, 'Sigmoid', 'Mul'], edges=[[0, 1], [1, 2], [0, 2]], exclusive=True
        )

        for pattern in patterns:
            computing, sigmoid, mul = pattern

            for var in sigmoid.outputs:
                self.graph.remove_variable(var)
            self.graph.remove_operation(sigmoid)

            input_vars = computing.outputs.copy()
            output_vars = mul.outputs.copy()

            self.graph.remove_operation(mul)
            self.graph.create_operation(
                op_type='Swish', inputs=input_vars, outputs=output_vars, name=f'{computing.name}/Swish'
            )
            assert len(input_vars) == 1, 'Fusion failed, Pattern unrecognized.'

    def fuse_bias_add(self):
        """
        Fuse Pattern like Conv + Add, ConvTranspose + Add, Gemm + Add
        This fusion will require a constant input as bias.
        """
        graph = self.graph
        for op in [_ for _ in graph.operations.values()]:
            if op.type in {'Conv', 'ConvTranspose', 'Gemm'}:
                # check if current op has only 1 downstream op
                channel_dimension = 1  # NCHW, NCHWD, NCH
                if op.type == 'Gemm':
                    # channel_dimension
                    pass
                if len(graph.get_downstream_operations(op)) == 1:
                    down = graph.get_downstream_operations(op)[0]

                    if down.type == 'Add':
                        if down.num_of_parameter != 1:
                            continue

                        bias = down.parameters[0]
                        if op.type not in {'Gemm'}:
                            # check if it is a bias add
                            if not bias.value.dim() == op.parameters[0].value.dim():
                                continue
                            if not bias.value.squeeze().dim() == 1:
                                continue
                            if bias.value.shape[channel_dimension] == 1:
                                continue
                            bias.value = bias.value.squeeze()  # conv bias can only be 1d
                        else:
                            # Gemm bias can be any shape.
                            # see https://github.com/onnx/onnx/blob/main/docs/Changelog.md#Gemm-11
                            pass

                        # ready for fusion
                        if op.num_of_input == 3:  # already has a bias
                            pass
                        else:
                            graph.create_variable(is_parameter=True, value=bias.value, dest_ops=[op])
                            graph.remove_operation(removing_op=down, keep_coherence=True)

    def fuse_scale(self):
        "Fuse Conv + Mul or Conv + Add"
        "Fuse ConvTranspose + Mul or ConvTranspose + Add"
        "Fuse Gemm + Mul or Gemm + Add"
        "Fuse MatMul + Mul"
        pass

    def fuse_selfattention(self):
        search_engine = SearchableGraph(graph=self.graph)
        matches = search_engine.pattern_matching(
            patterns=['MatMul', 'Add', 'Softmax', 'MatMul'], edges=[[0, 1], [1, 2], [2, 3]], exclusive=False
        )

        for m1, add, softmax, m2 in matches:
            trans_Q, trans_K, trans_V, trans_O = None, None, None, None
            perm_Q, perm_K, perm_V, perm_O = None, None, None, None

            # check pattern
            if m1.num_of_parameter != 0 or m2.num_of_parameter != 0:
                continue
            if m2.inputs[0].source_op != softmax:
                continue

            # source op of Q
            sq = m1.inputs[0].source_op
            if sq is not None and sq.type == 'Transpose':
                trans_Q = sq
                perm_Q = trans_Q.attributes['perm']

            # source op of K
            sk = m1.inputs[1].source_op
            if sk is not None and sq.type == 'Transpose':
                trans_K = sk
                perm_K = trans_K.attributes['perm']

            # source op of V
            sv = m2.inputs[1].source_op
            if sv is not None and sv.type == 'Transpose':
                trans_V = sv
                perm_V = trans_V.attributes['perm']

            # output op of O
            oo = m2.outputs[0].dest_ops
            if len(oo) == 1 and oo[0].type == 'Transpose':
                trans_O = oo[0]
                perm_O = trans_O.attributes['perm']

            for op in [trans_Q, trans_K, trans_V, trans_O, softmax]:
                if op is not None:
                    self.graph.remove_operation(op, keep_coherence=True)
            for var in [add.outputs[0], m1.outputs[0]]:
                self.graph.remove_variable(var)

            Q_var, K_var, V_var = m1.inputs[0], m1.inputs[1], m2.inputs[1]
            mask_var, O_var = add.inputs[0], m2.outputs[0]

            for op in m1, add, m2:
                self.graph.remove_operation(op)

            op = self.graph.create_operation(
                op_type='SelfAttention',
                attributes={'TransQ': perm_Q, 'TransK': perm_K, 'TransV': perm_V, 'TransO': perm_O},
                inputs=[Q_var, K_var, V_var, mask_var],
                outputs=[O_var],
            )

            # remove empty transpose
            non_empty_attr = {}
            for k, v in op.attributes.values():
                if v is not None:
                    non_empty_attr[k] = v
            op._attributes = non_empty_attr

        matches = search_engine.pattern_matching(
            patterns=['MatMul', 'Softmax', 'MatMul'], edges=[[0, 1], [1, 2]], exclusive=False
        )

        # self-attention with no mask
        for m1, softmax, m2 in matches:
            trans_Q, trans_K, trans_V, trans_O = None, None, None, None
            perm_Q, perm_K, perm_V, perm_O = None, None, None, None

            # check pattern
            if m1.num_of_parameter != 0 or m2.num_of_parameter != 0:
                continue
            if m2.inputs[0].source_op != softmax:
                continue

            # source op of Q
            sq = m1.inputs[0].source_op
            if sq is not None and sq.type == 'Transpose':
                trans_Q = sq
                perm_Q = trans_Q.attributes['perm']

            # source op of K
            sk = m1.inputs[1].source_op
            if sk is not None and sq.type == 'Transpose':
                trans_K = sk
                perm_K = trans_K.attributes['perm']

            # source op of V
            sv = m2.inputs[1].source_op
            if sv is not None and sv.type == 'Transpose':
                trans_V = sv
                perm_V = trans_V.attributes['perm']

            # output op of O
            oo = m2.outputs[0].dest_ops
            if len(oo) == 1 and oo[0].type == 'Transpose':
                trans_O = oo[0]
                perm_O = trans_O.attributes['perm']

            for op in [trans_Q, trans_K, trans_V, trans_O, softmax]:
                if op is not None:
                    self.graph.remove_operation(op, keep_coherence=True)
            for var in [m1.outputs[0]]:
                self.graph.remove_variable(var)

            Q_var, K_var, V_var = m1.inputs[0], m1.inputs[1], m2.inputs[1]
            O_var = m2.outputs[0]

            for op in m1, m2:
                self.graph.remove_operation(op)

            op = self.graph.create_operation(
                op_type='SelfAttention',
                attributes={'TransQ': perm_Q, 'TransK': perm_K, 'TransV': perm_V, 'TransO': perm_O},
                inputs=[Q_var, K_var, V_var, self.graph.create_variable()],
                outputs=[O_var],
            )

            # remove empty transpose
            non_empty_attr = {}
            for k, v in op.attributes.values():
                if v is not None:
                    non_empty_attr[k] = v
            op._attributes = non_empty_attr

    def fuse_matmul_add(self, verbose: bool = True):
        """
        Fuse Matmul + bias add to PPQBiasFusedMatMul

        PPQBiasFusedMatMul is a temporary operation which will be splited when exporting.
        """
        graph, fused = self.graph, False
        for current_op in [_ for _ in graph.operations.values()]:
            if current_op.type != 'MatMul':
                continue

            # check down-stream op is add
            next_ops = graph.get_downstream_operations(current_op)
            if len(next_ops) != 1:
                continue
            if next_ops[0].type != 'Add':
                continue

            # check if is a constant add
            fusing_op = next_ops[0]
            if fusing_op.num_of_parameter == 1:
                # do graph fusion
                bias = fusing_op.parameters[0].value
                graph.remove_operation(fusing_op, keep_coherence=True)
                graph.create_variable(value=bias, is_parameter=True, dest_ops=[current_op])
                current_op.type = 'PPQBiasFusedMatMul'
                fused = True

                if verbose:
                    print(f'Fusing graph op: {current_op.name} + {fusing_op.name}')

        if not fused:
            ppq_warning("No suitable matmul + add was found, check your graph again.")


class GraphDecomposer(GraphCommandProcessor):
    """Since PPQ 0.6.4, GraphDecomposer is introduced to split some complex
    operations For example, Gemm can be split with MatMul with Bias add.

    Gemm
    General Matrix multiplication: https://en.wikipedia.org/wiki/Basic_Linear_Algebra_Subprograms#Level_3

    A' = transpose(A) if transA else A

    B' = transpose(B) if transB else B

    Compute Y = alpha * A' * B' + beta * C, where input tensor A has shape (M, K) or (K, M),
        input tensor B has shape (K, N) or (N, K), input tensor C is broadcastable to shape (M, N),
        and output tensor Y has shape (M, N). A will be transposed before doing the computation if attribute transA is non-zero,
        same for B and transB.

    This operator supports unidirectional broadcasting (tensor C should be unidirectional broadcastable to tensor A * B);
        for more details please check the doc. This operator has optional inputs/outputs.

    See the doc for more details about the representation of optional arguments.
    An empty string may be used in the place of an actual argument's name to indicate a missing argument.
    Trailing optional arguments (those not followed by an argument that is present) may also be simply omitted.

    Attributes
        alpha : float (default is 1.0)
        Scalar multiplier for the product of input tensors A * B.

        beta : float (default is 1.0)
        Scalar multiplier for input tensor C.

        transA : int (default is 0)
        Whether A should be transposed

        transB : int (default is 0)
        Whether B should be transposed
    """

    def process(self, command: GraphCommand) -> Any:
        return super().process(command)

    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return super()._acceptable_command_types

    def decompose_gemm(self):
        graph = self.graph
        interested_ops = []
        for operation in graph.operations.values():
            if operation.type == 'Gemm':
                interested_ops.append(operation)

        for op in interested_ops:
            assert isinstance(op, Operation)
            output_var = op.outputs[0]

            if op.num_of_input == 3:
                bias_add = graph.create_operation(op_type='Add', platform=op.platform)
                bias_var = op.inputs[-1]

                graph.create_link_with_op(A=op, B=bias_add)
                graph.create_link_with_op(
                    variable=graph.create_variable(
                        value=bias_var.value * op.attributes.get('beta', 1), is_parameter=True
                    ),
                    A=None,
                    B=bias_add,
                )

                graph.remove_variable(bias_var)
                output_var.source_op = bias_add
                bias_add.outputs.append(output_var)
                op.outputs.remove(output_var)

            if op.attributes.get('transA', 0) == 1:
                raise ValueError(f'Can not process with operation {op.name}, transA=1 is not allowed.')
            if op.attributes.get('alpha', 1) != 1:
                op.parameters[0].value *= op.attributes.get('alpha')
            if op.attributes.get('transB', 0) == 1:
                op.inputs[1].value = op.inputs[1].value.permute(1, 0)

            op.type = 'MatMul'
            op.attributes.clear()

    def decompose_gru(self):
        pass

    def decompose_convtranspose(self):
        """Decompose ConvTranspose operation into InsertZeros + Conv.

        ConvTranspose(x, W, stride=s, padding=p, output_padding=op)
        is equivalent to:

        1. Insert zeros with stride s: x' = insert_zeros(x, stride=s)  # skip if stride=1
        2. Rotate kernel 180 degrees: W' = flip(W, dims=spatial_dims)
        3. Swap input/output channels: W'' = transpose(W', 0, 1)
        4. Conv with padding = kernel_size - p - 1: y = conv(x', W'', stride=1, padding=k-p-1)
        5. Add bias if present
        6. Apply output padding if needed

        This decomposition allows using standard Conv operation to implement ConvTranspose.
        Supports 1D, 2D, and 3D ConvTranspose.
        """
        graph = self.graph
        interested_ops = []
        for operation in graph.operations.values():
            if operation.type == 'ConvTranspose':
                interested_ops.append(operation)

        for op in interested_ops:
            assert isinstance(op, Operation)

            # Get ConvTranspose attributes
            strides = op.attributes.get('strides', [1, 1])
            pads_attr = op.attributes.get('pads', [0, 0, 0, 0])
            output_padding = op.attributes.get('output_padding', [0, 0])
            group = op.attributes.get('group', 1)
            dilations = op.attributes.get('dilations', [1, 1])
            kernel_shape_attr = op.attributes.get('kernel_shape', None)

            # Determine spatial dimensions from weight tensor
            weight_var = op.inputs[1]
            weight_value = weight_var.value
            weight_dim = weight_value.dim()
            spatial_dims = weight_dim - 2  # weight shape: [C_in, C_out, *spatial_kernel_dims]

            # Normalize attributes to lists of length spatial_dims
            def normalize_attr(attr, default, length, attr_name=""):
                if isinstance(attr, (int, float)):
                    return [attr] * length
                elif isinstance(attr, list):
                    if len(attr) == length:
                        return attr
                    elif len(attr) > length:
                        return attr[:length]
                    else:
                        raise ValueError(
                            f'ConvTranspose {op.name} normalize_attr error: {attr}, {default}, {length}, {attr_name}'
                        )
                else:
                    return [default] * length

            # Kernel shape from weight tensor is authoritative
            kernel_shape = list(weight_value.shape[2:])
            if kernel_shape_attr is not None:
                kernel_shape_attr = normalize_attr(kernel_shape_attr, 1, spatial_dims, "kernel")
                if kernel_shape_attr != kernel_shape:
                    ppq_warning(
                        f'ConvTranspose {op.name}: kernel_shape attribute {kernel_shape_attr} '
                        f'does not match weight shape {kernel_shape}. Using weight shape.'
                    )

            strides = normalize_attr(strides, 1, spatial_dims, "stride")
            # Special handling for pads
            # pads_attr is already obtained from op.attributes
            if pads_attr is None:
                # Default padding: 0 for all dimensions (start and end)
                pads_start = [0] * spatial_dims
                pads_end = [0] * spatial_dims
            elif isinstance(pads_attr, (int, float)):
                pads_start = [int(pads_attr)] * spatial_dims
                pads_end = [int(pads_attr)] * spatial_dims
            elif isinstance(pads_attr, list):
                if len(pads_attr) == spatial_dims:
                    # Only start padding provided, assume symmetric
                    pads_start = pads_attr
                    pads_end = pads_attr
                elif len(pads_attr) == 2 * spatial_dims:
                    # Full padding: [start_1, start_2, ..., end_1, end_2, ...]
                    pads_start = pads_attr[:spatial_dims]
                    pads_end = pads_attr[spatial_dims:]
                else:
                    # Try to extend or truncate
                    if len(pads_attr) > spatial_dims:
                        pads_start = pads_attr[:spatial_dims]
                        pads_end = pads_attr[:spatial_dims]  # assume symmetric
                    else:
                        pads_start = pads_attr + [0] * (spatial_dims - len(pads_attr))
                        pads_end = pads_start  # assume symmetric
            else:
                pads_start = [0] * spatial_dims
                pads_end = [0] * spatial_dims

            output_padding = normalize_attr(output_padding, 0, spatial_dims, "outut_padding")
            dilations = normalize_attr(dilations, 1, spatial_dims, "dilation")

            # Check if all strides are 1
            # Create InsertZeros operation if stride > 1
            all_strides_one = all(s == 1 for s in strides)
            need_insert_zeros = not all_strides_one

            if need_insert_zeros:
                # Create InsertZeros operation with stride and output_padding
                insert_zeros_op = graph.create_operation(
                    op_type='InsertZeros',
                    attributes={'stride': strides, 'output_padding': output_padding},
                    platform=op.platform,
                )

                # Link input to InsertZeros using graph helper
                input_var = op.inputs[0]
                graph.insert_op_before(A=insert_zeros_op, B=op, input_idx=0)

            # Calculate new padding for Conv: kernel_size - padding - 1
            # For asymmetric padding, compute start and end separately
            conv_pads_start = []
            conv_pads_end = []
            for k, p_start, p_end in zip(kernel_shape, pads_start, pads_end):
                pad_start = k - p_start - 1
                pad_end = k - p_end - 1
                conv_pads_start.append(pad_start)
                conv_pads_end.append(pad_end)
            conv_pads = conv_pads_start + conv_pads_end

            # Convert ConvTranspose to Conv operation
            op.type = 'Conv'
            op.attributes.clear()
            op.attributes['strides'] = [1] * spatial_dims
            op.attributes['pads'] = conv_pads
            op.attributes['dilations'] = dilations
            op.attributes['group'] = group
            op.attributes['kernel_shape'] = kernel_shape

            weight_value = weight_var.value.clone()  # Clone to avoid modifying original
            # Flip all spatial dimensions (dims 2 and above)
            spatial_dims_indices = list(range(2, weight_dim))
            weight_flipped = torch.flip(weight_value, dims=spatial_dims_indices)
            # Transpose input/output channels: ConvTranspose weight shape is [C_in, C_out, *spatial]
            # Conv weight shape is [C_out, C_in, *spatial]
            weight_for_conv = weight_flipped.transpose(0, 1).contiguous()

            # Update original weight variable for Conv
            weight_var.value = weight_for_conv


class GraphDeviceSwitcher(GraphCommandProcessor):
    """Graph Device Switcher insert necessary switcher operation for graph
    split and device dispatching.

    See also ppq scheduler for more information.

    All SOI operations are supposed to be executed on cpu.
        while other operations are supposed to be executed on cuda.
        Therefore switching operation will be inserted between SOI operations and fp32(quant) operations.
        to transfer cuda tensor to cpu tensor, vice versa.

    However some operations receive SOI input(cpu tensor) naturally, such as reshape, slice, etc.
    PPQ uses a tracing function for judging whether it is necessary to insert a
        switcher between operations like that.

    Before invoking this class, all operations must have been dispatched by a dispatcher.

    Args:
        GraphCommandProcessor ([type]): [description]
    """

    def insert_switcher(self):
        ppq_warning('This function has been removed from PPQ Since 0.6.6')

    def remove_switcher(self):
        ppq_warning('This function has been removed from PPQ Since 0.6.6')

    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return [GraphCommandType.INSERT_SWITCHER, GraphCommandType.REMOVE_SWITCHER]

    def process(self, command: GraphCommand) -> Any:
        if command.command_type == GraphCommandType.INSERT_SWITCHER:
            return self.insert_switcher()
        if command.command_type == GraphCommandType.REMOVE_SWITCHER:
            return self.remove_switcher()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Model(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Model()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsModel(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Model
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Model
    def IrVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Model
    def OpsetImport(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.OperatorSetId import OperatorSetId
            obj = OperatorSetId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Model
    def OpsetImportLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Model
    def OpsetImportIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Model
    def ProducerName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Model
    def ProducerVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Model
    def Domain(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Model
    def ModelVersion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # Model
    def DocString(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Model
    def Graph(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from FlatBuffers.Dl.Graph import Graph
            obj = Graph()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Model
    def MetadataProps(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.StringStringEntry import StringStringEntry
            obj = StringStringEntry()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Model
    def MetadataPropsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Model
    def MetadataPropsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        return o == 0

    # Model
    def Functions(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Function import Function
            obj = Function()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Model
    def FunctionsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Model
    def FunctionsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

def ModelStart(builder):
    builder.StartObject(10)

def Start(builder):
    ModelStart(builder)

def ModelAddIrVersion(builder, irVersion):
    builder.PrependInt32Slot(0, irVersion, 0)

def AddIrVersion(builder, irVersion):
    ModelAddIrVersion(builder, irVersion)

def ModelAddOpsetImport(builder, opsetImport):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(opsetImport), 0)

def AddOpsetImport(builder, opsetImport):
    ModelAddOpsetImport(builder, opsetImport)

def ModelStartOpsetImportVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartOpsetImportVector(builder, numElems):
    return ModelStartOpsetImportVector(builder, numElems)

def ModelAddProducerName(builder, producerName):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(producerName), 0)

def AddProducerName(builder, producerName):
    ModelAddProducerName(builder, producerName)

def ModelAddProducerVersion(builder, producerVersion):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(producerVersion), 0)

def AddProducerVersion(builder, producerVersion):
    ModelAddProducerVersion(builder, producerVersion)

def ModelAddDomain(builder, domain):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(domain), 0)

def AddDomain(builder, domain):
    ModelAddDomain(builder, domain)

def ModelAddModelVersion(builder, modelVersion):
    builder.PrependInt64Slot(5, modelVersion, 0)

def AddModelVersion(builder, modelVersion):
    ModelAddModelVersion(builder, modelVersion)

def ModelAddDocString(builder, docString):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(docString), 0)

def AddDocString(builder, docString):
    ModelAddDocString(builder, docString)

def ModelAddGraph(builder, graph):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(graph), 0)

def AddGraph(builder, graph):
    ModelAddGraph(builder, graph)

def ModelAddMetadataProps(builder, metadataProps):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(metadataProps), 0)

def AddMetadataProps(builder, metadataProps):
    ModelAddMetadataProps(builder, metadataProps)

def ModelStartMetadataPropsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartMetadataPropsVector(builder, numElems):
    return ModelStartMetadataPropsVector(builder, numElems)

def ModelAddFunctions(builder, functions):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(functions), 0)

def AddFunctions(builder, functions):
    ModelAddFunctions(builder, functions)

def ModelStartFunctionsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartFunctionsVector(builder, numElems):
    return ModelStartFunctionsVector(builder, numElems)

def ModelEnd(builder):
    return builder.EndObject()

def End(builder):
    return ModelEnd(builder)

import FlatBuffers.Dl.Function
import FlatBuffers.Dl.Graph
import FlatBuffers.Dl.OperatorSetId
import FlatBuffers.Dl.StringStringEntry
try:
    from typing import List, Optional
except:
    pass

class ModelT(object):

    # ModelT
    def __init__(self):
        self.irVersion = 0  # type: int
        self.opsetImport = None  # type: List[FlatBuffers.Dl.OperatorSetId.OperatorSetIdT]
        self.producerName = None  # type: str
        self.producerVersion = None  # type: str
        self.domain = None  # type: str
        self.modelVersion = 0  # type: int
        self.docString = None  # type: str
        self.graph = None  # type: Optional[FlatBuffers.Dl.Graph.GraphT]
        self.metadataProps = None  # type: List[FlatBuffers.Dl.StringStringEntry.StringStringEntryT]
        self.functions = None  # type: List[FlatBuffers.Dl.Function.FunctionT]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        model = Model()
        model.Init(buf, pos)
        return cls.InitFromObj(model)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, model):
        x = ModelT()
        x._UnPack(model)
        return x

    # ModelT
    def _UnPack(self, model):
        if model is None:
            return
        self.irVersion = model.IrVersion()
        if not model.OpsetImportIsNone():
            self.opsetImport = []
            for i in range(model.OpsetImportLength()):
                if model.OpsetImport(i) is None:
                    self.opsetImport.append(None)
                else:
                    operatorSetId_ = FlatBuffers.Dl.OperatorSetId.OperatorSetIdT.InitFromObj(model.OpsetImport(i))
                    self.opsetImport.append(operatorSetId_)
        self.producerName = model.ProducerName()
        self.producerVersion = model.ProducerVersion()
        self.domain = model.Domain()
        self.modelVersion = model.ModelVersion()
        self.docString = model.DocString()
        if model.Graph() is not None:
            self.graph = FlatBuffers.Dl.Graph.GraphT.InitFromObj(model.Graph())
        if not model.MetadataPropsIsNone():
            self.metadataProps = []
            for i in range(model.MetadataPropsLength()):
                if model.MetadataProps(i) is None:
                    self.metadataProps.append(None)
                else:
                    stringStringEntry_ = FlatBuffers.Dl.StringStringEntry.StringStringEntryT.InitFromObj(model.MetadataProps(i))
                    self.metadataProps.append(stringStringEntry_)
        if not model.FunctionsIsNone():
            self.functions = []
            for i in range(model.FunctionsLength()):
                if model.Functions(i) is None:
                    self.functions.append(None)
                else:
                    function_ = FlatBuffers.Dl.Function.FunctionT.InitFromObj(model.Functions(i))
                    self.functions.append(function_)

    # ModelT
    def Pack(self, builder):
        if self.opsetImport is not None:
            opsetImportlist = []
            for i in range(len(self.opsetImport)):
                opsetImportlist.append(self.opsetImport[i].Pack(builder))
            ModelStartOpsetImportVector(builder, len(self.opsetImport))
            for i in reversed(range(len(self.opsetImport))):
                builder.PrependUOffsetTRelative(opsetImportlist[i])
            opsetImport = builder.EndVector()
        if self.producerName is not None:
            producerName = builder.CreateString(self.producerName)
        if self.producerVersion is not None:
            producerVersion = builder.CreateString(self.producerVersion)
        if self.domain is not None:
            domain = builder.CreateString(self.domain)
        if self.docString is not None:
            docString = builder.CreateString(self.docString)
        if self.graph is not None:
            graph = self.graph.Pack(builder)
        if self.metadataProps is not None:
            metadataPropslist = []
            for i in range(len(self.metadataProps)):
                metadataPropslist.append(self.metadataProps[i].Pack(builder))
            ModelStartMetadataPropsVector(builder, len(self.metadataProps))
            for i in reversed(range(len(self.metadataProps))):
                builder.PrependUOffsetTRelative(metadataPropslist[i])
            metadataProps = builder.EndVector()
        if self.functions is not None:
            functionslist = []
            for i in range(len(self.functions)):
                functionslist.append(self.functions[i].Pack(builder))
            ModelStartFunctionsVector(builder, len(self.functions))
            for i in reversed(range(len(self.functions))):
                builder.PrependUOffsetTRelative(functionslist[i])
            functions = builder.EndVector()
        ModelStart(builder)
        ModelAddIrVersion(builder, self.irVersion)
        if self.opsetImport is not None:
            ModelAddOpsetImport(builder, opsetImport)
        if self.producerName is not None:
            ModelAddProducerName(builder, producerName)
        if self.producerVersion is not None:
            ModelAddProducerVersion(builder, producerVersion)
        if self.domain is not None:
            ModelAddDomain(builder, domain)
        ModelAddModelVersion(builder, self.modelVersion)
        if self.docString is not None:
            ModelAddDocString(builder, docString)
        if self.graph is not None:
            ModelAddGraph(builder, graph)
        if self.metadataProps is not None:
            ModelAddMetadataProps(builder, metadataProps)
        if self.functions is not None:
            ModelAddFunctions(builder, functions)
        model = ModelEnd(builder)
        return model



import numpy as np
import onnxruntime

# -------------------------------------------------------------------
# Onnxruntime 需要你提供一个 feed dict 和 output 的名字才能跑推理
# feed dict 就是 input name: data 的形式表示的输入数据
# output name 和 input name 你如果不知道的话，用可视化工具打开 onnx 文件就可以看到了。
# -------------------------------------------------------------------

MODEL = 'model.onnx'
FEED_DICT = {'input name': np.zeros(shape=[1, 3, 224, 224])}
OUTPUT_NAMES = ['output name']

session = onnxruntime.InferenceSession(MODEL, providers=['CUDAExecutionProvider'])
result = session.run(OUTPUT_NAMES, FEED_DICT)



import os
import sys
import warnings
from typing import Any, Dict, NamedTuple, Union, cast

import numpy as np

sys.path.append(os.path.dirname(__file__))
import FlatBuffers.Dl.OptionalType as OptionalType
import FlatBuffers.Dl.SequenceType as SequenceType
import FlatBuffers.Dl.Tensor
import FlatBuffers.Dl.Tensor as Tensor
import FlatBuffers.Dl.TensorDataType as TensorDataType


class TensorDtypeMap(NamedTuple):
    np_dtype: np.dtype
    storage_dtype: int
    name: str


# tensor_dtype: (numpy type, storage type, string name)
TENSOR_TYPE_MAP = {
    int(TensorDataType.TensorDataType().FLOAT): TensorDtypeMap(
        np.dtype("float32"), int(TensorDataType.TensorDataType().FLOAT), "TensorDataType.FLOAT"
    ),
    int(TensorDataType.TensorDataType().UINT8): TensorDtypeMap(
        np.dtype("uint8"), int(TensorDataType.TensorDataType().INT32), "TensorDataType.UINT8"
    ),
    int(TensorDataType.TensorDataType().INT8): TensorDtypeMap(
        np.dtype("int8"), int(TensorDataType.TensorDataType().INT32), "TensorDataType.INT8"
    ),
    int(TensorDataType.TensorDataType().UINT16): TensorDtypeMap(
        np.dtype("uint16"), int(TensorDataType.TensorDataType().INT32), "TensorDataType.UINT16"
    ),
    int(TensorDataType.TensorDataType().INT16): TensorDtypeMap(
        np.dtype("int16"), int(TensorDataType.TensorDataType().INT32), "TensorDataType.INT16"
    ),
    int(TensorDataType.TensorDataType().INT32): TensorDtypeMap(
        np.dtype("int32"), int(TensorDataType.TensorDataType().INT32), "TensorDataType.INT32"
    ),
    int(TensorDataType.TensorDataType().INT64): TensorDtypeMap(
        np.dtype("int64"), int(TensorDataType.TensorDataType().INT64), "TensorDataType.INT64"
    ),
    int(TensorDataType.TensorDataType().BOOL): TensorDtypeMap(
        np.dtype("bool"), int(TensorDataType.TensorDataType().INT32), "TensorDataType.BOOL"
    ),
    int(TensorDataType.TensorDataType().FLOAT16): TensorDtypeMap(
        np.dtype("float16"), int(TensorDataType.TensorDataType().UINT16), "TensorDataType.FLOAT16"
    ),
    int(TensorDataType.TensorDataType().DOUBLE): TensorDtypeMap(
        np.dtype("float64"), int(TensorDataType.TensorDataType().DOUBLE), "TensorDataType.DOUBLE"
    ),
    int(TensorDataType.TensorDataType().UINT32): TensorDtypeMap(
        np.dtype("uint32"), int(TensorDataType.TensorDataType().UINT32), "TensorDataType.UINT32"
    ),
    int(TensorDataType.TensorDataType().UINT64): TensorDtypeMap(
        np.dtype("uint64"), int(TensorDataType.TensorDataType().UINT64), "TensorDataType.UINT64"
    ),
    int(TensorDataType.TensorDataType().STRING): TensorDtypeMap(
        np.dtype("object"), int(TensorDataType.TensorDataType().STRING), "TensorDataType.STRING"
    ),
}


class DeprecatedWarningDict(dict):  # type: ignore
    def __init__(
        self,
        dictionary: Dict[int, Union[int, str, np.dtype]],
        original_function: str,
        future_function: str = "",
    ) -> None:
        super().__init__(dictionary)
        self._origin_function = original_function
        self._future_function = future_function

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DeprecatedWarningDict):
            return False
        return self._origin_function == other._origin_function and self._future_function == other._future_function

    def __getitem__(self, key: Union[int, str, np.dtype]) -> Any:
        if not self._future_function:
            warnings.warn(
                str(
                    f"`mapping.{self._origin_function}` is now deprecated and will be removed in a future release."
                    "To silence this warning, please simply use if-else statement to get the corresponding value."
                ),
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            warnings.warn(
                str(
                    f"`mapping.{self._origin_function}` is now deprecated and will be removed in a future release."
                    f"To silence this warning, please use `helper.{self._future_function}` instead."
                ),
                DeprecationWarning,
                stacklevel=2,
            )
        return super().__getitem__(key)


# This map is used for converting TensorProto values into numpy arrays
TENSOR_TYPE_TO_NP_TYPE = DeprecatedWarningDict(
    {tensor_dtype: value.np_dtype for tensor_dtype, value in TENSOR_TYPE_MAP.items()},
    "TENSOR_TYPE_TO_NP_TYPE",
    "tensor_dtype_to_np_dtype",
)


_NP_TYPE_TO_TENSOR_TYPE = {v: k for k, v in TENSOR_TYPE_TO_NP_TYPE.items()}


_STORAGE_TENSOR_TYPE_TO_FIELD = {
    int(TensorDataType.TensorDataType().FLOAT): Tensor.AddFloatData,
    int(TensorDataType.TensorDataType().INT32): Tensor.AddInt32Data,
    int(TensorDataType.TensorDataType().INT64): Tensor.AddInt64Data,
    int(TensorDataType.TensorDataType().UINT8): Tensor.AddInt32Data,
    int(TensorDataType.TensorDataType().UINT16): Tensor.AddInt32Data,
    int(TensorDataType.TensorDataType().DOUBLE): Tensor.AddDoubleData,
    int(TensorDataType.TensorDataType().UINT32): Tensor.AddUint64Data,
    int(TensorDataType.TensorDataType().UINT64): Tensor.AddUint64Data,
    int(TensorDataType.TensorDataType().STRING): Tensor.AddStringData,
    int(TensorDataType.TensorDataType().BOOL): Tensor.AddInt32Data,
}



import torch
from torch import nn
from torch.nn import functional as F


class Lenet5(nn.Module):
    """
    for cifar10 dataset.
    """

    def __init__(self):
        super(Lenet5, self).__init__()

        self.conv1 = nn.Conv2d(1, 6, kernel_size=5, stride=1, padding=0)
        self.pool1 = nn.AvgPool2d(kernel_size=2, stride=2, padding=0)
        self.conv2 = nn.Conv2d(6, 16, kernel_size=5, stride=1, padding=0)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        print('input: ', x.shape)
        x = F.relu(self.conv1(x))
        print('conv1', x.shape)
        x = self.pool1(x)
        print('pool1: ', x.shape)
        x = F.relu(self.conv2(x))
        print('conv2', x.shape)
        x = self.pool1(x)
        print('pool2', x.shape)
        x = x.view(x.size(0), -1)
        print('view: ', x.shape)
        x = F.relu(self.fc1(x))
        print('fc1: ', x.shape)
        x = F.relu(self.fc2(x))
        x = F.softmax(self.fc3(x), dim=1)
        return x


def export_onnx(onnx_filename):
    points = torch.full((1, 1, 32, 32), 1.5, dtype=torch.float32).cuda()
    inputs = (points,)
    model = Lenet5().cuda()
    torch.onnx.export(model, inputs, onnx_filename, opset_version=11)
    print("The output of raw network: ", torch.mean(model(*inputs)).detach().cpu().numpy())


def print_onnx_model(onnx_filename):
    # Load the ONNX model
    import onnx

    model = onnx.load(onnx_filename)
    # Check that the IR is well formed
    onnx.checker.check_model(model)
    # Print a human readable representation of the graph
    print(onnx.helper.printable_graph(model.graph))


def main():
    # get onnx
    onnx_filename = "lenet.onnx"
    export_onnx(onnx_filename)
    print_onnx_model(onnx_filename)


if __name__ == '__main__':
    main()



import json
from typing import List

import torch

from esp_ppq.core import DataType, NetworkFramework, QuantizationStates, QuantizationVisibility, ppq_warning
from esp_ppq.IR import BaseGraph, GraphExporter
from esp_ppq.IR.quantize import QuantableOperation

from .onnx_exporter import OnnxExporter
from .util import convert_value


class QNNDSPExporter(GraphExporter):
    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        activation_info, param_info = {}, {}
        topo_order = graph.topological_sort()
        for operation in topo_order:
            if not isinstance(operation, QuantableOperation):
                continue
            for config, var in operation.config_with_variable:
                if not QuantizationStates.can_export(config.state):
                    raise PermissionError(
                        'Can not export quant config cause not all quantization configurations '
                        'have been correctly initialized(or some of them has been deactivated). '
                        f'Operation {operation.name} has an invalid quantization state({config.state}) '
                        f'at variable {var.name}.'
                    )

                if config.visibility == QuantizationVisibility.INTERNAL:
                    continue
                if config.state in {QuantizationStates.FP32, QuantizationStates.SOI}:
                    continue

                if var.source_op is not None and var.source_op.type in {"Softmax", "Sigmoid"}:
                    if config.dominated_by == config:  # changeable.
                        # fix output range greater than 1
                        config.scale = torch.clamp(config.scale, max=1.0 / (config.quant_max - config.quant_min))

                if config.state == QuantizationStates.PASSIVE and var.name in activation_info:
                    continue
                info = [
                    {
                        'bitwidth': config.num_of_bits,
                        'max': convert_value(config.scale * (config.quant_max - config.offset), True, DataType.FP32),
                        'min': convert_value(config.scale * (config.quant_min - config.offset), True, DataType.FP32),
                        'offset': convert_value(config.offset, True, DataType.INT32),
                        'scale': convert_value(config.scale, True, DataType.FP32),
                    }
                ]
                if var.is_parameter:
                    param_info[var.name] = info
                else:
                    activation_info[var.name] = info

        exports = {'activation_encodings': activation_info, 'param_encodings': param_info}

        with open(file=config_path, mode='w') as file:
            json.dump(exports, file, indent=4)

    def export(
        self,
        file_path: str,
        graph: BaseGraph,
        config_path: str = None,
        input_shapes: List[List[int]] = [[1, 3, 224, 224]],
    ):
        if config_path is not None:
            self.export_quantization_config(config_path, graph)
        if graph._built_from == NetworkFramework.ONNX:
            exporter = OnnxExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None)



from abc import ABCMeta, abstractmethod
from typing import Callable, Dict, List, Union

import torch

from esp_ppq.core import TargetPlatform, TensorQuantizationConfig
from esp_ppq.executor.op import (
    DEFAULT_BACKEND_TABLE,
    ESPDL_QUANT_BACKEND_TABLE,
    EXTENSION_BACKEND_TABLE,
    NXP_BACKEND_TABLE,
    PPL_DSP_BACKEND_TABLE,
    PPL_GPU_BACKEND_TABLE,
)
from esp_ppq.IR import BaseGraph, Operation, QuantableOperation

OPERATION_FORWARD_TABLE = {platform: DEFAULT_BACKEND_TABLE.copy() for platform in TargetPlatform}

OPERATION_FORWARD_TABLE[TargetPlatform.TRT_INT8] = PPL_GPU_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.PPL_CUDA_INT8] = PPL_GPU_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.PPL_CUDA_INT4] = PPL_GPU_BACKEND_TABLE

OPERATION_FORWARD_TABLE[TargetPlatform.PPL_DSP_INT8] = PPL_DSP_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.PPL_DSP_TI_INT8] = PPL_DSP_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.NXP_INT8] = NXP_BACKEND_TABLE

OPERATION_FORWARD_TABLE[TargetPlatform.EXTENSION] = EXTENSION_BACKEND_TABLE

OPERATION_FORWARD_TABLE[TargetPlatform.ESPDL_INT8] = ESPDL_QUANT_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.ESPDL_INT16] = ESPDL_QUANT_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.ESPDL_S3_H_PRE_INT16] = ESPDL_QUANT_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.ESPDL_S3_INT16] = ESPDL_QUANT_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.ESPDL_S3_INT8] = ESPDL_QUANT_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.ESPDL_C_INT8] = ESPDL_QUANT_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.ESPDL_C_INT16] = ESPDL_QUANT_BACKEND_TABLE
OPERATION_FORWARD_TABLE[TargetPlatform.ESPDL_C_H_PRE_INT16] = ESPDL_QUANT_BACKEND_TABLE


def register_operation_handler(handler: Callable, operation_type: str, platform: TargetPlatform):
    """Regitser a custimized function as operation handler.

    Function should accept at least 3 input parameters, return one or more tensor as result:
    func(op: Operation, values: List[torch.Tensor], ctx: TorchBackendContext = None, **kwargs) -> torch.Tensor:

    If there is already another operation handler for given operation_type,
        new handler will replace the old one without warrning.

    Args:
        handler (Callable): Callable function, which interface follows restriction above.
        operation_type (str): Operation type.
        platform (TargetPlatform): Register platform.
    """
    if platform not in OPERATION_FORWARD_TABLE:
        raise ValueError('Unknown Platform detected, Please check your platform setting.')
    OPERATION_FORWARD_TABLE[platform][operation_type] = handler


class RuntimeHook(metaclass=ABCMeta):
    """RuntimeHook is an abstract class designed for executor customizing.

    Args:
        metaclass ([type], optional): [description]. Defaults to ABCMeta.
    """

    def __init__(self, operation: Operation, **kwargs) -> None:
        self._hook_to = operation

    def pre_forward_hook(self, inputs: list, **kwargs) -> list:
        """user-customized pre-processing procedure of input data.

        Args:
            inputs (list): a list includes all input data.

        Returns:
            list: a list includes all input data(processed).
        """
        return inputs

    def post_forward_hook(self, outputs: list, **kwargs) -> list:
        """user-customized post-processing procedure of output data.

        Args:
            inputs (list): a list includes all output data.

        Returns:
            list: a list includes all output data(processed).
        """
        return outputs


class QuantOPRuntimeHook(RuntimeHook, metaclass=ABCMeta):
    """QuantOPRuntimeHook is an abstract class designed for executor
    customizing.

    Args:
        metaclass ([type], optional): [description]. Defaults to ABCMeta.
    """

    def __init__(self, operation: QuantableOperation, **kwargs) -> None:
        if not isinstance(operation, QuantableOperation):
            raise TypeError(f'You are trying to bind a QuantRuntimeHook to a non-quantized operation {operation}.')
        super().__init__(operation)

    def pre_forward_hook(self, inputs: list, quant_inputs: list, quant_configs: List[TensorQuantizationConfig]) -> list:
        return quant_inputs

    def post_forward_hook(
        self, outputs: list, quant_outputs: list, quant_configs: List[TensorQuantizationConfig]
    ) -> list:
        return quant_outputs


class BaseGraphExecutor(Callable, metaclass=ABCMeta):
    """PPQ Base Graph Executor.

    Args:
        Callable ([type]): [description]
        metaclass ([type], optional): [description]. Defaults to ABCMeta.
    """

    def __init__(self, graph: BaseGraph) -> dict:
        self._graph = None
        self._graph_input_dictionary = None
        self._graph_output_dictionary = None
        self._executing_order = None
        self.load_graph(graph=graph)

    def load_graph(self, graph: BaseGraph) -> dict:
        self._graph = graph
        self._graph_input_dictionary = self._graph.inputs
        self._graph_output_dictionary = self._graph.outputs
        self._executing_order = self._graph.topological_sort()

    def prepare_input(self, inputs: Union[dict, list, torch.Tensor]):
        assert type(inputs) in (dict, list, torch.Tensor), (
            f'Input format misunderstood. Except either dict, list or tensor; while {type(inputs)} was given.'
        )

        inputs_dictionary = self._graph.inputs
        if len(inputs_dictionary) == 0:
            assert inputs is None, 'Graph do not need any inputs. please set your inputs to be None.'
            return None

        if isinstance(inputs, torch.Tensor):
            assert len(inputs_dictionary) == 1, 'Graph needs more than one input, while only one tensor was given.'
            return {list(inputs_dictionary.keys())[0]: inputs}

        elif isinstance(inputs, list):
            assert len(inputs_dictionary) == len(inputs), (
                f'Inputs format misunderstood. Given inputs has '
                f'{len(inputs)} elements, while graph needs {len(inputs_dictionary)}'
            )
            return {key: inputs[idx] for idx, key in enumerate(inputs_dictionary)}

        elif isinstance(inputs, dict):
            assert len(inputs_dictionary) == len(inputs), (
                f'Inputs format misunderstood. Given inputs has '
                f'{len(inputs)} elements, while graph needs {len(inputs_dictionary)}'
            )
            return inputs

        else:
            raise Exception('Oops, you can never reach here.')

    @abstractmethod
    def forward(
        self,
        inputs: Union[dict, list, torch.Tensor],
        output_names: List[str] = None,
        hooks: Dict[str, RuntimeHook] = None,
    ) -> List[torch.Tensor]:
        raise NotImplementedError('Please implement this function first.')

    @abstractmethod
    def tracing_operation_meta(self, inputs: Union[dict, list, torch.Tensor], output_names: List[str] = None) -> None:
        raise NotImplementedError('Please implement this function first.')

    def __call__(self, inputs: Union[dict, torch.Tensor], output_names: List[str] = None) -> List[torch.Tensor]:
        return self.forward(inputs=inputs, output_names=output_names)

    def __str__(self) -> str:
        return f'PPQ GraphExecuter Object: {self.__hash__()}'



from math import ceil
from typing import Callable, Dict, Iterable, List

import torch
from tqdm import tqdm

from esp_ppq.core import (
    OBSERVER_ISOTONE_OBSERVER_AXIS,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
    empty_ppq_cache,
)
from esp_ppq.executor import BaseGraphExecutor, RuntimeHook
from esp_ppq.IR import BaseGraph, QuantableOperation, QuantableVariable
from esp_ppq.quantization.observer import (
    CalibrationHook,
    OperationObserver,
    TensorObserverFactroy,
    TorchHistObserver,
    TorchMinMaxObserver,
    TorchMSEObserver,
)

from .base import QuantizationOptimizationPass


class RuntimeCalibrationPass(QuantizationOptimizationPass):
    """
    ## Runtime Calibration Pass(量化参数校准过程)

    For integer quantization, you need to calibrate or estimate the scale of all floating-point tensors in the model.

    Formula:

            Quant(Y, scale_Y) = Clip(Round(Y / scale_Y))

            Dequant(Y, scale_Y) = Y * scale_Y

    Only activations that have quantization state = INITIAL are going to be calibrated via this optimization pass.
    While if the parameter "override" is set to True, activations with quantization state = ACTIVATED will also be re-calibrated.

    Runtime Calibration Pass will write estimated scales and offsets to tensor quantization configs, and set their state to ACTIVATED.

    Unlike constant tensors such as weights and biases, variable tensors such as model input,
    activations (outputs of intermediate layers) and model output cannot be calibrated unless we run a few inference cycles.

    As a result, PPQ Runtime Calibration Pass requires a representative dataset to calibrate them.

    This dataset is supposed to be a small subset (around ~100-500 samples) of the training or validation data.

    ### Parameters:

    * method(str):

            String that representing the algorithm used to estimate scales and offsets for activations.

            Can be mse, kl, percentile, minmax, this parameter is case insensitive.

            You can register your own calibration method through functions in esp_ppq.api

    * override(bool)

            if this parameter is set to True, activations with quantization state = ACTIVATED will also be re-calibrated,
            runtime calibration pass will overwrite their scales and offsets.

            This parameter is introduced since ppq 0.6.4

    ### observer support matrix:

    | observer     | Symmetrical | Asymmetrical | Per-chanel | Per-tensor | Cuda Acceleration   |
    | ---          | ---         | ---          | ---        | ---        |                     |
    | minmax       | [x]         | [x]          | [x]        | [x]        | [ ]                 |
    | mse          | [x]         | [x]          | [ ]        | [x]        | [x]                 |
    | precentile   | [x]         | [x]          | [x]        | [x]        | [x]                 |
    | kl           | [x]         | [ ]          | [ ]        | [x]        | [x]                 |
    | isotone      | [x]         | [x]          | [ ]        | [x]        | [ ]                 |

    ### Usage:

    Runtime Calibration Pass should be invoked before Passive Parameter Quantize Pass

    This pass is included in PPQ Quantization Setting, you can calling this optimization by:

        setting = QuantizationSettingFactory.default_setting()

        setting.quantize_activation = True

        # calling esp_ppq.api.quantize_onnx_model function with this setting.
        ir = quantize_torch_model(
        model=model, calib_dataloader=load_calibration_dataset(), setting=setting,
        platform=TargetPlatform.PPL_CUDA_INT8, calib_steps=8, input_shape=INPUT_SHAPE,
        collate_fn=collate_fn)

    You can manually create this optimization by:

        from esp_ppq import RuntimeCalibrationPass

        optim = RuntimeCalibrationPass()

    ### Register Calibration Method:

    Using api function register_calibration_observer to resister new observer algorithm to PPQ system.
    Once Algorithm is registered, Runtime Calibration Pass will automatically calling them by name.

    This feature requires PPQ > 0.6.5

    """

    def __init__(self, method: str = None, override: bool = False, calib_steps: int = 32) -> None:
        super().__init__(name='PPQ Runtime Calibration Pass')
        self._method = method
        self._observers = {}
        self._collate_fn = None
        self._calib_steps = calib_steps
        self._override = override

    def calibrate(
        self,
        desc: str,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        hooks: Dict[str, RuntimeHook],
        output_names: List[str] = None,
    ):
        calib_step = 0
        with tqdm(total=self._calib_steps, desc=desc) as progressing_bar:
            for calib_epoch in range(ceil(self._calib_steps / len(dataloader))):
                for data in dataloader:
                    if self._collate_fn is not None:
                        data = self._collate_fn(data)
                    executor.forward(inputs=data, hooks=hooks, output_names=output_names)
                    progressing_bar.update()
                    calib_step += 1
                    if calib_step >= self._calib_steps:
                        break

    @empty_ppq_cache
    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        calib_steps: int = 32,
        collate_fn: Callable = None,
        **kwargs,
    ) -> None:
        if collate_fn is not None:
            self._collate_fn = collate_fn
        if calib_steps is not None:
            self._calib_steps = calib_steps

        assert calib_steps >= 2, (
            'Insufficient Calibration Detected, to get a better quantization performance, '
            'more calibration steps is required, we strongly recommend you to prepare more calibration data '
            'and more calibration steps is preferred here. (at least 8)'
        )

        assert calib_steps <= 5000, (
            'Calibration steps is too large, ppq can quantize your network within 8-5000 '
            'calibration steps. More calibration steps will greatly delay ppq\'s calibration procedure. '
            'Reset your calib_steps parameter please.'
        )

        # -------------------------------------------------
        # Override existing quantization configurations
        # -------------------------------------------------
        if self._override:
            for operation in graph.operations.values():
                if not isinstance(operation, QuantableOperation):
                    continue

                for config, var in operation.config_with_variable:
                    if (
                        not var.is_parameter
                        and config.state == QuantizationStates.ACTIVATED
                        and config.dominated_by == config
                    ):
                        config.state = QuantizationStates.INITIAL

        # build observer and hook for each quantable operation
        hooks = {}
        for op_name, operation in graph.operations.items():
            if not isinstance(operation, QuantableOperation):
                continue

            # override algorithm setting if necessary
            for config, var in operation.config_with_variable:
                if not var.is_parameter and self._method is not None:
                    config.observer_algorithm = self._method

            observer = OperationObserver(operation=executor._graph.operations[op_name], monitor_parameter=False)
            self._observers[op_name] = observer
            hooks[op_name] = observer.hook

        # ready for calibration
        # hook forward function, let observers take effects.
        self.calibrate(
            desc='Calibration Progress(Phase 1)',
            dataloader=dataloader,
            executor=executor,
            hooks=hooks,
            output_names=None,
        )

        # render calibration result.
        for _, observer in self._observers.items():
            assert isinstance(observer, OperationObserver)
            observer.render_quantization_config()
            observer.report()

        # -------------------------------------------------
        # There are some two-phase observer in ppq,
        # which means they have to be calibrated for a second time.
        #   see also: TorchHistObserver
        # -------------------------------------------------

        # remove one-phase observer from hook dict.
        pop_list = []
        for op_name, observer in self._observers.items():
            assert isinstance(observer, OperationObserver)
            if all(
                [
                    type(var_observer) not in {TorchHistObserver, TorchMSEObserver}
                    for var_observer in observer._hook._observer_table.values()
                ]
            ):
                pop_list.append(op_name)

        for op_name in pop_list:
            self._observers.pop(op_name)
            hooks.pop(op_name)

        if len(hooks) > 0:
            # ready for calibration(Phase 2)
            # hook forward function, let observers take effects.
            self.calibrate(
                desc='Calibration Progress(Phase 2)',
                dataloader=dataloader,
                executor=executor,
                hooks=hooks,
                output_names=None,
            )

            # render calibration result for a second time.
            for _, observer in self._observers.items():
                assert isinstance(observer, OperationObserver)
                observer.render_quantization_config()
                observer.report()


class PPLDSPTIReCalibrationPass(RuntimeCalibrationPass):
    """PPQ ReCalibration Pass For Computing Ops This pass should only be turned
    on when the platform is one of PPL DSP TI series, which needs a per-channel
    recalibration process for output variable of computing op types.

    This pass does not interfere with the normal quantization process, and will
    be turned      off for most situations.
    """

    def __init__(self, method: str = None, override: bool = False) -> None:
        super().__init__(method, override)
        self.name = 'PPQ ReCalibration For Computing Op Pass'

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        calib_steps: int,
        collate_fn: Callable,
        **kwargs,
    ) -> None:
        self._collate_fn = collate_fn
        self._calib_steps = calib_steps
        assert calib_steps >= 8, (
            'Insufficient Calibration Detected, to better quantize your network, '
            'more calibration steps is demonded, we strongly recommend you to prepare more calibration data '
            'and more calibration steps is preferred here. (at least 8)'
        )

        assert calib_steps <= 512, (
            'Calibration steps is too large, ppq is capable for quantizing your network within 32-128 '
            'calibration steps. More calibraiton steps will greatly delay ppq\'s calibration procedure. '
            'Reset your calib_steps parameter please.'
        )

        hooks = {}
        for operation in tqdm(graph.topological_sort(), desc='Collecting Observer For Computing Ops'):
            if not isinstance(operation, QuantableOperation) or not operation.is_computing_op:
                continue
            output_cfg = operation.config.output_quantization_config[0]
            master_cfg, master_operation, master_var = output_cfg, operation, operation.outputs[0]

            observe_table = {}
            # to check if all input data is greater than 0
            # we only need a basic observer here
            if operation.inputs[0].name in graph.inputs:
                input_cfg = operation.config.input_quantization_config[0]
                sym_input_cfg = TensorQuantizationConfig(
                    policy=QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
                    ),
                    rounding=input_cfg.rounding,
                    num_of_bits=input_cfg.num_of_bits,
                    quant_min=input_cfg.quant_min,
                    quant_max=input_cfg.quant_max,
                    scale=None,
                    offset=None,
                    observer_algorithm='Minmax',
                    detail={'consumer': input_cfg},
                )
                observe_table.update(
                    {input_cfg: TensorObserverFactroy.build_observer(operation.inputs[0], sym_input_cfg)}
                )

            # only consider overlapped by relu/clip activation here
            downstream_ops = graph.get_downstream_operations(operation)
            if (
                len(downstream_ops) == 1
                and downstream_ops[0].type in {'Relu', 'Clip'}
                and isinstance(downstream_ops[0], QuantableOperation)
            ):
                if len(observe_table) > 0:
                    hooks[operation.name] = CalibrationHook(operation, observe_table)
                    observe_table = {}
                master_cfg = downstream_ops[0].config.output_quantization_config[0]
                master_operation = downstream_ops[0]
                master_var = downstream_ops[0].outputs[0]

            master_cfg_per_channel = TensorQuantizationConfig(
                policy=QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                ),
                rounding=master_cfg.rounding,
                num_of_bits=master_cfg.num_of_bits,
                quant_min=master_cfg.quant_min,
                quant_max=master_cfg.quant_max,
                scale=None,
                offset=None,
                observer_algorithm='Minmax',
                state=QuantizationStates.INITIAL,
                channel_axis=1,
                detail={'consumer': output_cfg},
            )

            observe_table.update({master_cfg: TensorObserverFactroy.build_observer(master_var, master_cfg_per_channel)})
            assert master_operation.name not in hooks, 'register an operation in calibration hooks twice'
            hooks[master_operation.name] = CalibrationHook(master_operation, observe_table)

        self.calibrate(desc='ReCalibration For Computing Ops', dataloader=dataloader, executor=executor, hooks=hooks)

        for hook in hooks.values():
            assert isinstance(hook, CalibrationHook)
            # hook.render_quantization_config()
            for observer in hook._observer_table.values():
                cfg = observer._quant_cfg.detail['consumer']
                assert isinstance(cfg, TensorQuantizationConfig)
                assert isinstance(observer, TorchMinMaxObserver)

                if observer._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                    min_vals = (
                        torch.min(torch.cat(observer._min_val_collector, dim=-1), dim=-1, keepdim=False)[0]
                        .cpu()
                        .numpy()
                    )
                    max_vals = (
                        torch.max(torch.cat(observer._max_val_collector, dim=-1), dim=-1, keepdim=False)[0]
                        .cpu()
                        .numpy()
                    )
                    cfg.detail.update({'range_min': min_vals, 'range_max': max_vals})

                elif observer._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                    min_val = (torch.min(torch.cat(observer._min_val_collector, dim=0)).cpu().item(),)
                    max_val = (torch.max(torch.cat(observer._max_val_collector, dim=0)).cpu().item(),)
                    cfg.detail.update({'range_min': min_val, 'range_max': max_val})


class IsotoneCalibrationPass(RuntimeCalibrationPass):
    """
    ## Isotone Calibration Pass(保序量化校准过程)

    在神经网络中，一些算子的输出并不需要保证总体的精确性，而只关注于最大最小值所在的位置，
    例如图像分类网络中，网络的输出通常是一个1000维的向量，用于表达图像属于特定类别的概率。
    为了保证分类的正确性，我们并不需要这个1000维的向量在量化后是整体准确的，只需要其中的最大值出现在正确的位置上。
    因此我们希望最大值与次大值之间相差至少半个 scale，并且次大值能够不被截断。

    因此传统的 min-max, percentile, kl 方法在这一情景中并不能得到最高的分类精度，
    保序量化是为了解决这一问题而设计的，在这一校准过程中，程序将网络输出变量的校准方式改写为 Isotone(保序校准)。
    默认设置下，该过程只对 softmax 算子的输出进行保序校准。对于其他情况，用户需要手动指定需要进行保序校准的变量名。

    保序量化需要设定一个分类轴，同样地以分类网络为例，其输出形为 [Batch, 1000]。
    分类操作将在数据的最后一维展开，因此需要设置保序轴为 -1。

    Algorithm:

        For softmax or sigmoid activations, usually we just need
        argmax(softmax(x)) == argmax(softmax(quant(x)))

        Inspired by this Property, Isotone Observer is designed to provide an order-preserving calibration method,
            which cares only about argmax(x) [or argmin(x)]

        To keep argmax(x) == argmax(quant(x)), we only need to
            distinguish the largest element and the second largert element with quantization

            let L1 represents the largest element of x,
            while L2 represents the second largest.

            For Symmetrical Quantization, We want:

                1. round(L1 / scale) - round(L2 / scale) > 0

                2. round(L2 / scale) < quant_max

            Hence that, we will have:

                1. scale < 2 * (L1 - L2)

                2. scale > L2 / (self._quant_cfg.quant_max - .5)

            For Asymmetircal Quantization, We want:

                1. round(L1 / scale) + offset - round(L2 / scale) - offset > 0

                2. round(L2 / scale) + offset < quant_max

            Hence that, we will have:

                1. scale < 2 * (L1 - L2)

                2. scale > L2 / (self._quant_cfg.quant_max - offset - .5)

        The best setting of scale, offset can be solved by PPQ Isotone observer.

        Time Complexity: O(nlogn)
    """

    def __init__(
        self, variables: List[str] = None, axis: int = -1, verbose: bool = True, calib_steps: int = 32
    ) -> None:
        super().__init__(calib_steps=calib_steps)
        self.name = "Isotone Calibration Pass"
        self.variables = variables
        self.axis = axis
        self.verbose = verbose

    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        if self.variables is None:
            for op in graph.operations.values():
                if op.type == 'Softmax' and isinstance(op, QuantableOperation):
                    # had not been dominated.
                    if op.output_quant_config[0].dominated_by == op.output_quant_config[0]:
                        op.output_quant_config[0].state = QuantizationStates.INITIAL
                        op.output_quant_config[0].observer_algorithm = 'Isotone'
                        op.output_quant_config[0].detail[OBSERVER_ISOTONE_OBSERVER_AXIS] = op.attributes.get('axis', -1)

                        if self.verbose:
                            print(
                                f'Calibration Method of Op {op.name} '
                                f'has been changed to Isotone[axis={op.attributes.get("axis", -1)}].'
                            )

        else:  # self.variables is not None
            if not isinstance(self.variables, list):
                raise TypeError('Isotone Calibration Pass needs a list of variable name as its input.')
            for var in self.variables:
                if not isinstance(var, str):
                    raise TypeError('Isotone Calibration Pass needs a list of variable name as its input.')
                if var not in graph.variables:
                    raise ValueError(f'Variable {var} not in current graph.')

                var = graph.variables[var]
                if isinstance(var, QuantableVariable):
                    var.source_op_config.state = QuantizationStates.INITIAL
                    var.source_op_config.observer_algorithm = 'Isotone'
                    var.source_op_config.detail[OBSERVER_ISOTONE_OBSERVER_AXIS] = self.axis
                    if self.verbose:
                        print(
                            f'Calibration Method of Variable {var.name} has been changed to Isotone[axis={self.axis}].'
                        )

        super().optimize(graph=graph, **kwargs)


class QuantConfigModifyPass(QuantizationOptimizationPass):
    """
    Set quantization scale of selected layers
    Note: only modify the first output by default
    """

    def __init__(
        self,
        custom_config: Dict[str, float],
        verbose: bool = True,
    ) -> None:
        super().__init__(name='PPQ Quant Config Modify Pass')
        self.custom_config = custom_config
        self.verbose = verbose

    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        for layer_name in self.custom_config.keys():
            if layer_name not in graph.operations:
                raise KeyError(f'Operation {layer_name} not found in graph.')

            op = graph.operations[layer_name]

            if not isinstance(op, QuantableOperation):
                raise TypeError(f'Operation {layer_name} is not QuantableOperation')

            out_cfg = op.output_quant_config[0]

            if out_cfg.dominated_by is not out_cfg:
                try:
                    dom_name = out_cfg.dominated_by.operation.name
                except AttributeError:
                    dom_name = "UnknownOp"
                if self.verbose:
                    print(f'[Skip] {layer_name} output is dominated by {dom_name}')
                continue

            # calibration config
            out_cfg.state = QuantizationStates.INITIAL

            # scale
            out_cfg.scale = torch.tensor([self.custom_config[layer_name]], dtype=torch.float32).squeeze(0)
            out_cfg.offset = torch.tensor([0], dtype=torch.float32).squeeze(0)
            out_cfg.state = QuantizationStates.ACTIVATED

            if self.verbose:
                print(
                    f'[Focused Quantization]Set OP {layer_name} output quantization scale to {self.custom_config[layer_name]}'
                )



from typing import Union

import torch

from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.executor.op.torch.base import GET_ATTRIBUTE_FROM_OPERATION
from esp_ppq.IR import BaseGraph, Operation

from .base import BaseQuantizer


def ASSERT_CONV_AND_DECONV(op: Operation):
    if op.type == "Conv":
        filter = op.inputs[1].value.shape
        assert len(filter) == 4, (
            f'Ascend Quantization needs the dimension of filter must be 4, while your filter is {len(filter)}'
        )

    if op.type == "ConvTranspose":
        filter = op.inputs[1].value.shape
        dilation = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='dilations', compulsive=True, default=1)
        groups = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='group', compulsive=True, default=1)

        assert len(filter) == 4, (
            f'Ascend Quantization needs the dimension of filter must be 4, while your filter is {len(filter)}'
        )
        assert dilation == [1, 1], (
            f'Ascend Quantization needs the dilation must be [1, 1], while your dilation is {dilation}'
        )
        assert groups == 1, f'Ascend Quantization needs the groups must be 1, while your groups is {groups}'


def ASSERT_GEMM(op: Operation):
    axis = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='axis', compulsive=True, default=1)
    assert axis == 1, f'Ascend Quantization needs the axis must be 1, while your axis is {axis}'

    if 'transpose' in op.attributes:
        transpose = op.attributes['transpose']
        assert transpose == False, (
            f'Ascend Quantization needs the transpose must be False, while your transpose is {transpose}'
        )


class AscendQuantizer(BaseQuantizer):
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = 0
        self._quant_max = int(pow(2, self._num_of_bits) - 1)

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
        )

        ASSERT_CONV_AND_DECONV(operation)

        if operation.type in {'Conv', 'Gemm', 'ConvTranspose'}:
            assert operation.num_of_input > 0, 'Seems you got a Computing layer with no parameters.'

            if operation.type in {'Conv', 'ConvTranspose'}:
                if operation.inputs[1].is_parameter:
                    conv_weight_config = base_quant_config.input_quantization_config[1]
                    conv_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL
                        + QuantizationProperty.LINEAR
                        + QuantizationProperty.PER_CHANNEL
                    )
                    conv_weight_config.channel_axis = 1 if operation.type == 'ConvTranspose' else 0
                    conv_weight_config.observer_algorithm = 'minmax'
                    conv_weight_config.quant_max = 127
                    conv_weight_config.quant_min = -128

            elif operation.type == 'Gemm':
                ASSERT_GEMM(operation)

                if operation.inputs[1].is_parameter:
                    gemm_weight_config = base_quant_config.input_quantization_config[1]
                    gemm_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
                    )

                    # gemm_weight_config.channel_axis = 0
                    gemm_weight_config.observer_algorithm = 'minmax'
                    gemm_weight_config.quant_max = 127
                    gemm_weight_config.quant_min = -128

            # if operation has bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False
        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ASC_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {'Conv', 'ConvTranspose', 'Gemm', 'AveragePool'}

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.ASYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}



from .default import DEFAULT_BACKEND_TABLE

ESPDL_QUANT_BACKEND_TABLE = DEFAULT_BACKEND_TABLE.copy()

from esp_ppq.core import (
    GRU_QUANT_BITS,
    GRU_QUANT_EXPONENT,
    LSTM_QUANT_BITS,
    LSTM_QUANT_EXPONENT,
    RoundingPolicy,
    TargetPlatform,
    TensorQuantizationConfig,
)
from esp_ppq.quantization.qfunction import PPQuantFunction
from esp_ppq.utils.round import ppq_tensor_round

from .base import *


def GRU_float_forward(
    op: Operation, values: List[torch.Tensor], ctx: TorchBackendContext = None, **kwargs
) -> torch.Tensor:
    """Computes an one-layer GRU using basic PyTorch operations."""
    ASSERT_NUM_OF_INPUT(op=op, values=values, min_num_of_input=3, max_num_of_input=6)
    values = VALUE_TO_EXECUTING_DEVICE(op=op, ctx=ctx, values=values)

    # Extract inputs
    x, w, r = values[:3]
    b = GET_VALUE_FROM_INPUTS(values, 3)
    seq_len = GET_VALUE_FROM_INPUTS(values, 4)
    initial_h = GET_VALUE_FROM_INPUTS(values, 5)

    # Get attributes
    hidden_size = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='hidden_size', compulsive=True)
    direction = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='direction', default='forward')
    layout = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='layout', default=0)

    # Configuration
    bidirectional = direction == 'bidirectional'
    num_directions = 2 if bidirectional else 1
    batch_first = layout == 1
    has_bias = b is not None

    # Reshape input if batch_first
    if batch_first:
        x = x.transpose(0, 1)  # [batch, seq, input] -> [seq, batch, input]

    seq_length, batch_size, input_size = x.shape

    # Initialize hidden state
    if initial_h is None:
        h = torch.zeros(num_directions, batch_size, hidden_size, device=x.device, dtype=x.dtype)
    else:
        h = initial_h

    # Prepare biases according to GRU formula
    def prepare_biases(bias, hidden_size, direction_idx=0):
        if bias is None:
            return (None, None, None, None, None, None)

        if bias.dim() == 2:  # [num_directions, ...]
            b_dir = bias[direction_idx]
        else:
            b_dir = bias

        # Split into Wb (input bias) and Rb (recurrent bias) components
        # Bias layout: [Wb_z, Wb_r, Wb_h, Rb_z, Rb_r, Rb_h]
        wb_z = b_dir[:hidden_size]  # Wbz
        wb_r = b_dir[hidden_size : 2 * hidden_size]  # Wbr
        wb_h = b_dir[2 * hidden_size : 3 * hidden_size]  # Wbh
        rb_z = b_dir[3 * hidden_size : 4 * hidden_size]  # Rbz
        rb_r = b_dir[4 * hidden_size : 5 * hidden_size]  # Rbr
        rb_h = b_dir[5 * hidden_size : 6 * hidden_size]  # Rbh

        return wb_z, wb_r, wb_h, rb_z, rb_r, rb_h

    # Process single direction
    def process_direction(x, w, r, b, initial_h, direction_idx=0, reverse=False):
        # Get weights for this direction
        if w.dim() == 3:  # [num_directions, 3*hidden_size, input_size]
            w_dir = w[direction_idx]  # [3*hidden_size, input_size]
            r_dir = r[direction_idx]  # [3*hidden_size, hidden_size]
        else:
            w_dir = w  # [3*hidden_size, input_size]
            r_dir = r  # [3*hidden_size, hidden_size]

        # Get biases for this direction
        if has_bias:
            wb_z, wb_r, wb_h, rb_z, rb_r, rb_h = prepare_biases(b, hidden_size, direction_idx)
        else:
            wb_z, wb_r, wb_h, rb_z, rb_r, rb_h = None, None, None, None, None, None

        # Initialize hidden state for this direction
        h_t = initial_h[direction_idx] if initial_h.dim() == 3 else initial_h

        # Reverse sequence if needed
        if reverse:
            x = x.flip(0)

        # Process sequence
        outputs = []
        for t in range(seq_length):
            x_t = x[t]  # [batch_size, input_size]

            # Compute all gates at once: x_t @ W^T -> [batch_size, 3*hidden_size]
            xw = torch.matmul(x_t, w_dir.t())  # [batch_size, 3*hidden_size]

            # Split into z, r, h components
            xw_z, xw_r, xw_h = torch.split(xw, hidden_size, dim=1)

            # Compute all recurrent connections at once: h_t @ R^T -> [batch_size, 3*hidden_size]
            hr = torch.matmul(h_t, r_dir.t())  # [batch_size, 3*hidden_size]

            # Split into z, r, h components
            hr_z, hr_r, hr_h = torch.split(hr, hidden_size, dim=1)

            # Update gate (z): zt = σ(Xt*(Wz^T) + Ht-1*(Rz^T) + Wbz + Rbz)
            z_input = xw_z + hr_z
            if wb_z is not None:
                z_input = z_input + wb_z + rb_z  # Wbz
            z_t = torch.sigmoid(z_input)

            # Reset gate (r): rt = σ(Xt*(Wr^T) + Ht-1*(Rr^T) + Wbr + Rbr)
            r_input = xw_r + hr_r
            if wb_r is not None:
                r_input = r_input + wb_r + rb_r  # Wbr
            r_t = torch.sigmoid(r_input)

            # Hidden gate (h): ht = tanh(Xt*(Wh^T) + (rt ⊙ (Ht-1*(Rh^T) + Rbh)) + Wbh)
            # linear_before_reset != 0 的情况

            # 计算 Ht-1*(Rh^T) + Rbh
            hr_h_with_bias = hr_h
            if rb_h is not None:
                hr_h_with_bias = hr_h_with_bias + rb_h  # Rbh

            # 计算 rt ⊙ (Ht-1*(Rh^T) + Rbh)
            reset_gated_hidden = r_t * hr_h_with_bias

            # 计算 Xt*(Wh^T) + Wbh
            xw_h_with_bias = xw_h
            if wb_h is not None:
                xw_h_with_bias = xw_h_with_bias + wb_h  # Wbh

            # 最终隐藏门计算
            h_input = xw_h_with_bias + reset_gated_hidden
            h_tilde = torch.tanh(h_input)

            # New hidden state: Ht = (1 - zt) ⊙ ht + zt ⊙ Ht-1
            h_t = (1 - z_t) * h_tilde + z_t * h_t
            outputs.append(h_t)

        outputs = torch.stack(outputs)  # [seq_length, batch_size, hidden_size]

        if reverse:
            outputs = outputs.flip(0)

        return outputs, h_t

    # Process forward direction
    forward_output, forward_final = process_direction(x, w, r, b, h, 0, False)

    if bidirectional:
        # Process reverse direction
        reverse_output, reverse_final = process_direction(x, w, r, b, h, 1, True)

        # Concatenate outputs
        outputs = torch.cat([forward_output.unsqueeze(1), reverse_output.unsqueeze(1)], dim=1)
        final_hidden = torch.stack([forward_final, reverse_final])
    else:
        outputs = forward_output.unsqueeze(1)  # [seq_length, 1, batch_size, hidden_size]
        final_hidden = forward_final.unsqueeze(0)  # [1, batch_size, hidden_size]

    return outputs, final_hidden


def fake_quantize(
    tensor,
    exponent=GRU_QUANT_EXPONENT,
    bits=GRU_QUANT_BITS,
    rounding: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN,
):
    scale = pow(2, exponent)
    quant_min = -pow(2, bits - 1)
    quant_max = -quant_min - 1
    tensor = ppq_tensor_round((tensor / scale), rounding)
    tensor = torch.clamp(tensor, quant_min, quant_max)
    tensor = tensor * scale
    return tensor


def GRU_quant_forward(
    op: Operation,
    values: List[torch.Tensor],
    ctx: TorchBackendContext = None,
    **kwargs,
) -> torch.Tensor:
    """Computes an one-layer GRU using basic PyTorch operations."""
    ASSERT_NUM_OF_INPUT(op=op, values=values, min_num_of_input=3, max_num_of_input=6)
    values = VALUE_TO_EXECUTING_DEVICE(op=op, ctx=ctx, values=values)

    # Extract inputs
    x, w, r = values[:3]
    b = GET_VALUE_FROM_INPUTS(values, 3)
    seq_len = GET_VALUE_FROM_INPUTS(values, 4)
    initial_h = GET_VALUE_FROM_INPUTS(values, 5)

    # Get attributes
    hidden_size = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='hidden_size', compulsive=True)
    direction = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='direction', default='forward')
    layout = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='layout', default=0)

    # Configuration
    bidirectional = direction == 'bidirectional'
    num_directions = 2 if bidirectional else 1
    batch_first = layout == 1
    has_bias = b is not None

    # Reshape input if batch_first
    if batch_first:
        x = x.transpose(0, 1)  # [batch, seq, input] -> [seq, batch, input]

    seq_length, batch_size, input_size = x.shape

    # Initialize hidden state
    if initial_h is None:
        h = torch.zeros(num_directions, batch_size, hidden_size, device=x.device, dtype=x.dtype)
    else:
        h = initial_h

    # Prepare biases according to GRU formula
    def prepare_biases(bias, hidden_size, direction_idx=0):
        if bias is None:
            return (None, None, None, None, None, None)

        if bias.dim() == 2:  # [num_directions, ...]
            b_dir = bias[direction_idx]
        else:
            b_dir = bias

        # Split into Wb (input bias) and Rb (recurrent bias) components
        # Bias layout: [Wb_z, Wb_r, Wb_h, Rb_z, Rb_r, Rb_h]
        wb_z = b_dir[:hidden_size]  # Wbz
        wb_r = b_dir[hidden_size : 2 * hidden_size]  # Wbr
        wb_h = b_dir[2 * hidden_size : 3 * hidden_size]  # Wbh
        rb_z = b_dir[3 * hidden_size : 4 * hidden_size]  # Rbz
        rb_r = b_dir[4 * hidden_size : 5 * hidden_size]  # Rbr
        rb_h = b_dir[5 * hidden_size : 6 * hidden_size]  # Rbh

        z_bias = wb_z + rb_z
        # z_bias = fake_quantize(z_bias)
        r_bias = wb_r + rb_r
        # r_bias = fake_quantize(r_bias)

        return z_bias, r_bias, wb_h, rb_h

    # Process single direction
    def process_direction(
        x, w, r, b, initial_h, direction_idx=0, reverse=False, quant_config: TensorQuantizationConfig = None
    ):
        # Get weights for this direction
        if w.dim() == 3:  # [num_directions, 3*hidden_size, input_size]
            w_dir = w[direction_idx]  # [3*hidden_size, input_size]
            r_dir = r[direction_idx]  # [3*hidden_size, hidden_size]
        else:
            w_dir = w  # [3*hidden_size, input_size]
            r_dir = r  # [3*hidden_size, hidden_size]

        if quant_config is not None:
            rounding = quant_config.rounding
        else:
            rounding = RoundingPolicy.ROUND_HALF_EVEN

        # Get biases for this direction
        if has_bias:
            z_bias, r_bias, wb_h, rb_h = prepare_biases(b, hidden_size, direction_idx)
        else:
            z_bias, r_bias, wb_h, rb_h = None, None, None, None

        # Initialize hidden state for this direction
        h_t = initial_h[direction_idx] if initial_h.dim() == 3 else initial_h

        # Reverse sequence if needed
        if reverse:
            x = x.flip(0)

        # Process sequence
        outputs = []
        for t in range(seq_length):
            x_t = x[t]  # [batch_size, input_size]

            # Compute all gates at once: x_t @ W^T -> [batch_size, 3*hidden_size]
            xw = torch.matmul(x_t, w_dir.t())  # [batch_size, 3*hidden_size]
            xw = fake_quantize(xw, GRU_QUANT_EXPONENT, GRU_QUANT_BITS, rounding=rounding)

            # Split into z, r, h components
            xw_z, xw_r, xw_h = torch.split(xw, hidden_size, dim=1)

            # Compute all recurrent connections at once: h_t @ R^T -> [batch_size, 3*hidden_size]
            hr = torch.matmul(h_t, r_dir.t())  # [batch_size, 3*hidden_size]
            hr = fake_quantize(hr, GRU_QUANT_EXPONENT, GRU_QUANT_BITS, rounding=rounding)

            # Split into z, r, h components
            hr_z, hr_r, hr_h = torch.split(hr, hidden_size, dim=1)

            # Update gate (z): zt = σ(Xt*(Wz^T) + Ht-1*(Rz^T) + Wbz + Rbz)
            z_input = xw_z + hr_z
            if z_bias is not None:
                z_input = z_input + z_bias  # Wbz
            z_t = torch.sigmoid(z_input)

            # Reset gate (r): rt = σ(Xt*(Wr^T) + Ht-1*(Rr^T) + Wbr + Rbr)
            r_input = xw_r + hr_r
            if r_bias is not None:
                r_input = r_input + r_bias  # Wbr
            r_t = torch.sigmoid(r_input)

            # Hidden gate (h): ht = tanh(Xt*(Wh^T) + (rt ⊙ (Ht-1*(Rh^T) + Rbh)) + Wbh)
            # linear_before_reset != 0 的情况

            # 计算 Ht-1*(Rh^T) + Rbh
            hr_h_with_bias = hr_h
            if rb_h is not None:
                hr_h_with_bias = hr_h_with_bias + rb_h  # Rbh

            # 计算 rt ⊙ (Ht-1*(Rh^T) + Rbh)
            reset_gated_hidden = r_t * hr_h_with_bias

            # 计算 Xt*(Wh^T) + Wbh
            xw_h_with_bias = xw_h
            if wb_h is not None:
                xw_h_with_bias = xw_h_with_bias + wb_h  # Wbh

            # 最终隐藏门计算
            h_input = fake_quantize(
                xw_h_with_bias + reset_gated_hidden, GRU_QUANT_EXPONENT, GRU_QUANT_BITS, rounding=rounding
            )
            h_tilde = torch.tanh(h_input)

            # New hidden state: Ht = (1 - zt) ⊙ ht + zt ⊙ Ht-1
            h_t = (1 - z_t) * h_tilde + z_t * h_t

            # scale is not None when model is quantized
            h_t = PPQuantFunction(h_t, quant_config)
            outputs.append(h_t)

        outputs = torch.stack(outputs)  # [seq_length, batch_size, hidden_size]

        if reverse:
            outputs = outputs.flip(0)

        return outputs, h_t

    # Create quantization config for gates and hidden state

    if len(op.config.output_quantization_config) > 0:
        output_config = op.config.output_quantization_config[0]
    else:
        raise TypeError('GRU_quant_forward except a TensorQuantizationConfig instance.')

    # Process forward direction
    forward_output, forward_final = process_direction(x, w, r, b, h, 0, False, output_config)

    if bidirectional:
        # Process reverse direction
        reverse_output, reverse_final = process_direction(x, w, r, b, h, 1, True, output_config)

        # Concatenate outputs
        outputs = torch.cat([forward_output.unsqueeze(1), reverse_output.unsqueeze(1)], dim=1)
        final_hidden = torch.stack([forward_final, reverse_final])
    else:
        outputs = forward_output.unsqueeze(1)  # [seq_length, 1, batch_size, hidden_size]
        final_hidden = forward_final.unsqueeze(0)  # [1, batch_size, hidden_size]

    return outputs, final_hidden


# -------------- LSTM --------------


def LSTM_float_forward(
    op: Operation,
    values: List[torch.Tensor],
    ctx=None,
    **kwargs,
) -> torch.Tensor:
    ASSERT_NUM_OF_INPUT(op=op, values=values, min_num_of_input=3, max_num_of_input=8)
    values = VALUE_TO_EXECUTING_DEVICE(op=op, ctx=ctx, values=values)

    x, w, r = values[:3]  # x: [seq, batch, input]  (layout=0)
    b = GET_VALUE_FROM_INPUTS(values, 3)
    seq_len = GET_VALUE_FROM_INPUTS(values, 4)
    initial_h = GET_VALUE_FROM_INPUTS(values, 5)
    initial_c = GET_VALUE_FROM_INPUTS(values, 6)
    p = GET_VALUE_FROM_INPUTS(values, 7)  # peephole weight, 这里先留接口，暂不使用

    hidden_size = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='hidden_size', compulsive=True)
    direction = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='direction', default='forward')
    layout = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='layout', default=0)
    bidirectional = direction == 'bidirectional'
    num_directions = 2 if bidirectional else 1
    batch_first = layout == 1
    has_bias = b is not None

    if batch_first:
        x = x.transpose(0, 1)  # -> [seq, batch, input]

    seq_length, batch_size, input_size = x.shape

    # --- 初始化 h0/c0 ---
    if initial_h is None:
        initial_h = torch.zeros(num_directions, batch_size, hidden_size, device=x.device, dtype=x.dtype)
    if initial_c is None:
        initial_c = torch.zeros_like(initial_h)

    # --- 拆 bias：LSTM 有 8 个 bias 向量 (Wb_i,f,g,o + Rb_i,f,g,o) ---
    def prepare_lstm_biases(bias, hidden_size, direction_idx=0):
        if bias is None:
            return (None,) * 8
        if bias.dim() == 2:
            b_dir = bias[direction_idx]  # [8*hidden]
        else:
            b_dir = bias
        splits = torch.split(b_dir, hidden_size, dim=0)  # 8 份
        return splits  # (Wb_i, Wb_f, Wb_g, Wb_o, Rb_i, Rb_f, Rb_g, Rb_o)

    # --- 单方向计算 ---
    def process_direction(x, w, r, b, h0, c0, direction_idx=0, reverse=False):
        # 取该方向权重
        if w.dim() == 3:
            w_dir = w[direction_idx]  # [4*hidden, input]
            r_dir = r[direction_idx]  # [4*hidden, hidden]
        else:
            w_dir, r_dir = w, r

        # - W/R/B parameter weight matrix for input, output, forget, and cell gates
        # 处理 bias
        if has_bias:
            Wb_i, Wb_o, Wb_f, Wb_g, Rb_i, Rb_o, Rb_f, Rb_g = prepare_lstm_biases(b, hidden_size, direction_idx)
            # 合并成 gate_bias = Wb + Rb
            i_bias = Wb_i + Rb_i
            f_bias = Wb_f + Rb_f
            g_bias = Wb_g + Rb_g
            o_bias = Wb_o + Rb_o
        else:
            i_bias = f_bias = g_bias = o_bias = None

        h_t = h0[direction_idx] if h0.dim() == 3 else h0
        c_t = c0[direction_idx] if c0.dim() == 3 else c0

        if reverse:
            x = x.flip(0)

        outputs = []
        for t in range(seq_length):
            xt = x[t]  # [batch, input]

            # 一次 matmul 得到 4 个门
            xw = torch.matmul(xt, w_dir.t())  # [batch, 4*hidden]
            x_i, x_o, x_f, x_g = torch.split(xw, hidden_size, dim=1)

            hr = torch.matmul(h_t, r_dir.t())  # [batch, 4*hidden]
            h_i, h_o, h_f, h_g = torch.split(hr, hidden_size, dim=1)

            # 加 bias
            if has_bias:
                x_i, x_f, x_g, x_o = x_i + i_bias, x_f + f_bias, x_g + g_bias, x_o + o_bias

            i_t = torch.sigmoid(x_i + h_i)
            f_t = torch.sigmoid(x_f + h_f)
            g_t = torch.tanh(x_g + h_g)
            o_t = torch.sigmoid(x_o + h_o)

            # 更新 cell
            c_t = f_t * c_t + i_t * g_t
            h_t = o_t * torch.tanh(c_t)

            # 对输出 h_t 再做一次量化（与 GRU 侧保持一致）
            outputs.append(h_t)

        outputs = torch.stack(outputs)  # [seq, batch, hidden]
        if reverse:
            outputs = outputs.flip(0)
        return outputs, h_t, c_t

    # --- 前向 ---
    fo, fh, fc = process_direction(x, w, r, b, initial_h, initial_c, 0, False)
    if bidirectional:
        ro, rh, rc = process_direction(x, w, r, b, initial_h, initial_c, 1, True)
        outputs = torch.cat([fo.unsqueeze(1), ro.unsqueeze(1)], dim=1)  # [seq, 2, batch, hidden]
        last_h = torch.stack([fh, rh])  # [2, batch, hidden]
        last_c = torch.stack([fc, rc])
    else:
        outputs = fo.unsqueeze(1)  # [seq, 1, batch, hidden]
        last_h = fh.unsqueeze(0)
        last_c = fc.unsqueeze(0)

    return outputs, last_h, last_c


def LSTM_quant_forward(
    op: Operation,
    values: List[torch.Tensor],
    ctx=None,
    **kwargs,
) -> torch.Tensor:
    ASSERT_NUM_OF_INPUT(op=op, values=values, min_num_of_input=3, max_num_of_input=8)
    values = VALUE_TO_EXECUTING_DEVICE(op=op, ctx=ctx, values=values)

    x, w, r = values[:3]  # x: [seq, batch, input]  (layout=0)
    b = GET_VALUE_FROM_INPUTS(values, 3)
    seq_len = GET_VALUE_FROM_INPUTS(values, 4)
    initial_h = GET_VALUE_FROM_INPUTS(values, 5)
    initial_c = GET_VALUE_FROM_INPUTS(values, 6)
    p = GET_VALUE_FROM_INPUTS(values, 7)  # peephole weight, 这里先留接口，暂不使用

    hidden_size = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='hidden_size', compulsive=True)
    direction = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='direction', default='forward')
    layout = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='layout', default=0)
    bidirectional = direction == 'bidirectional'
    num_directions = 2 if bidirectional else 1
    batch_first = layout == 1
    has_bias = b is not None

    if batch_first:
        x = x.transpose(0, 1)  # -> [seq, batch, input]

    seq_length, batch_size, input_size = x.shape

    # --- 初始化 h0/c0 ---
    if initial_h is None:
        initial_h = torch.zeros(num_directions, batch_size, hidden_size, device=x.device, dtype=x.dtype)
    if initial_c is None:
        initial_c = torch.zeros_like(initial_h)

    # --- 拆 bias：LSTM 有 8 个 bias 向量 (Wb_i,f,g,o + Rb_i,f,g,o) ---
    def prepare_lstm_biases(bias, hidden_size, direction_idx=0):
        if bias is None:
            return (None,) * 8
        if bias.dim() == 2:
            b_dir = bias[direction_idx]  # [8*hidden]
        else:
            b_dir = bias
        splits = torch.split(b_dir, hidden_size, dim=0)  # 8 份
        return splits  # (Wb_i, Wb_f, Wb_g, Wb_o, Rb_i, Rb_f, Rb_g, Rb_o)

    # --- 单方向计算 ---
    def process_direction(
        x, w, r, b, h0, c0, direction_idx=0, reverse=False, quant_config: TensorQuantizationConfig = None
    ):
        # 取该方向权重
        if w.dim() == 3:
            w_dir = w[direction_idx]  # [4*hidden, input]
            r_dir = r[direction_idx]  # [4*hidden, hidden]
        else:
            w_dir, r_dir = w, r

        rounding = quant_config.rounding if quant_config else RoundingPolicy.ROUND_HALF_EVEN

        # 处理 bias
        if has_bias:
            Wb_i, Wb_o, Wb_f, Wb_g, Rb_i, Rb_o, Rb_f, Rb_g = prepare_lstm_biases(b, hidden_size, direction_idx)
            # 合并成 gate_bias = Wb + Rb
            i_bias = Wb_i + Rb_i
            f_bias = Wb_f + Rb_f
            g_bias = Wb_g + Rb_g
            o_bias = Wb_o + Rb_o
        else:
            i_bias = f_bias = g_bias = o_bias = None

        h_t = h0[direction_idx] if h0.dim() == 3 else h0
        c_t = c0[direction_idx] if c0.dim() == 3 else c0

        if reverse:
            x = x.flip(0)

        outputs = []
        for t in range(seq_length):
            xt = x[t]  # [batch, input]

            # 一次 matmul 得到 4 个门
            xw = torch.matmul(xt, w_dir.t())  # [batch, 4*hidden]
            xw = fake_quantize(xw, LSTM_QUANT_EXPONENT, LSTM_QUANT_BITS, rounding=rounding)
            x_i, x_o, x_f, x_g = torch.split(xw, hidden_size, dim=1)

            hr = torch.matmul(h_t, r_dir.t())  # [batch, 4*hidden]
            hr = fake_quantize(hr, LSTM_QUANT_EXPONENT, LSTM_QUANT_BITS, rounding=rounding)
            h_i, h_o, h_f, h_g = torch.split(hr, hidden_size, dim=1)

            # 加 bias
            if has_bias:
                x_i, x_f, x_g, x_o = x_i + i_bias, x_f + f_bias, x_g + g_bias, x_o + o_bias

            i_t = torch.sigmoid(x_i + h_i)
            f_t = torch.sigmoid(x_f + h_f)
            g_t = torch.tanh(x_g + h_g)
            o_t = torch.sigmoid(x_o + h_o)

            # 更新 cell
            c_t = f_t * c_t + i_t * g_t
            c_t = fake_quantize(c_t, LSTM_QUANT_EXPONENT, LSTM_QUANT_BITS, rounding=rounding)  # 很多框架把 c_t 也量化
            h_t = o_t * torch.tanh(c_t)

            # 对输出 h_t 再做一次量化（与 GRU 侧保持一致）
            h_t = PPQuantFunction(h_t, quant_config)

            outputs.append(h_t)

        outputs = torch.stack(outputs)  # [seq, batch, hidden]
        if reverse:
            outputs = outputs.flip(0)
        return outputs, h_t, c_t

    # --- 取输出量化配置 ---
    if len(op.config.output_quantization_config) > 0:
        output_config = op.config.output_quantization_config[0]
    else:
        raise TypeError('LSTM_quant_forward expects a TensorQuantizationConfig instance.')

    # --- 前向 ---
    fo, fh, fc = process_direction(x, w, r, b, initial_h, initial_c, 0, False, output_config)
    if bidirectional:
        ro, rh, rc = process_direction(x, w, r, b, initial_h, initial_c, 1, True, output_config)
        outputs = torch.cat([fo.unsqueeze(1), ro.unsqueeze(1)], dim=1)  # [seq, 2, batch, hidden]
        last_h = torch.stack([fh, rh])  # [2, batch, hidden]
        last_c = torch.stack([fc, rc])
    else:
        outputs = fo.unsqueeze(1)  # [seq, 1, batch, hidden]
        last_h = fh.unsqueeze(0)
        last_c = fc.unsqueeze(0)

    return outputs, last_h, last_c


ESPDL_QUANT_BACKEND_TABLE['GRU'] = GRU_quant_forward
ESPDL_QUANT_BACKEND_TABLE['LSTM'] = LSTM_quant_forward



# This example shows how to make a dynamic-shape network
# dynamic shape is only supported by onnx

# first of all, load your model from anywhere
from esp_ppq import *
from esp_ppq.api import *

YOU_WANT_TO_QUANTIZE_IT = True

ir = load_onnx_graph('onnx model path')
input_shape = [1, 3, 224, 224]
samples = [torch.zeros(size=input_shape).cuda()]

if YOU_WANT_TO_QUANTIZE_IT:
    ir = dispatch_graph(ir, platform=TargetPlatform.NCNN_INT8, setting=QuantizationSettingFactory.ncnn_setting())

    ir = quantize_native_model(
        model=ir,
        calib_dataloader=samples,
        calib_steps=32,
        input_shape=input_shape,
        setting=QuantizationSettingFactory.ncnn_setting(),
    )

# You are supposed to set dynamic shape input/output variable just before export.
# Get variable instance from ir by its name, set shape attribute as your wish.
var = ir.variables['input variable name']
var.shape = ['Batch', 3, 'Width', 'Height']
# text, None, int are both acceptable here.

export_ppq_graph(
    graph=ir, platform=TargetPlatform.NCNN_INT8, graph_save_to='onnx model save to', config_save_to='config save to'
)



from typing import List


def shape_to_str(shape: List[int]) -> str:
    if len(shape) == 1:
        return str(shape[0])
    string_builder = str(shape[0])
    for s in shape[1:]:
        string_builder += '_' + str(s)
    return string_builder



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class AttributeF(object):
    __slots__ = ['_tab']

    @classmethod
    def SizeOf(cls):
        return 4

    # AttributeF
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # AttributeF
    def F(self): return self._tab.Get(flatbuffers.number_types.Float32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))

def CreateAttributeF(builder, f):
    builder.Prep(4, 4)
    builder.PrependFloat32(f)
    return builder.Offset()


class AttributeFT(object):

    # AttributeFT
    def __init__(self):
        self.f = 0.0  # type: float

    @classmethod
    def InitFromBuf(cls, buf, pos):
        attributeF = AttributeF()
        attributeF.Init(buf, pos)
        return cls.InitFromObj(attributeF)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, attributeF):
        x = AttributeFT()
        x._UnPack(attributeF)
        return x

    # AttributeFT
    def _UnPack(self, attributeF):
        if attributeF is None:
            return
        self.f = attributeF.F()

    # AttributeFT
    def Pack(self, builder):
        return CreateAttributeF(builder, self.f)



from .cosine import numpy_cosine_similarity, torch_cosine_similarity, torch_cosine_similarity_as_loss
from .norm import torch_mean_square_error, torch_snr_error
from .statistic import torch_KL_divergence
